/*
* bos.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] = "$Id: bos.cc,v 1.20 2002/08/17 15:27:53 root Exp $";

/*
 * $Log: bos.cc,v $
 * Revision 1.20  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.19  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.18  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.17  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.16  2001/12/16 17:56:08  root
 * #
 *
 * Revision 1.15  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.14  2001/12/16 17:42:43  root
 * #
 *
 * Revision 1.13  2001/12/16 17:41:11  root
 * #
 *
 * Revision 1.12  2001/12/16 17:39:17  root
 * #
 *
 * Revision 1.11  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.10  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.9  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.8  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.7  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.6  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.5  2001/12/06 04:26:39  root
 * #
 *
 */

#include "project.h"

#include "coords/libcoords.h"

#define CORRECT_X 0
#define CORRECT_Y 0


/* Utility functions*/

static ImageCoord
bosxytobic (BOSXY xy)
{
  return ImageCoord (xy.x, xy.y);
}

static BOSXY
bictobosxy (ImageCoord ic)
{
  return BOSXY (ic.x, ic.y);
}

static WGS84
lltowgs84 (LatLon ll)
{
  return WGS84 (ll.lat, ll.lon);
}

static LatLon
wgs84toll (WGS84 w)
{
  return LatLon (w.lat, w.lon);
}

/* BOS Image coordinate system */

string BosImgCoordSystem::name (void)
{
  string
  ret;
  char
  buf[1024];
  sprintf (buf, "British Ordinance Survey grid");
  ret = buf;
  return ret;
}

ImageCoord BosImgCoordSystem::lltoic (LatLon ll)
{
  BOSXY
  xy =
    wgs84_to_bosxy (lltowgs84 (ll));
  return ImageCoord (bosxytobic (xy));
}

LatLon BosImgCoordSystem::ictoll (ImageCoord ic)
{
  WGS84
  w =
    bosxy_to_wgs84 (bictobosxy (ic));
  return LatLon (w.lat, w.lon);
}

/* Utility functions to handle those wacky letters*/

static int
letter_to_x (char c)
{
  switch (c | 32) {
    case 'a':
    case 'f':
    case 'l':
    case 'q':
    case 'v':
      return 0;
    case 'b':
    case 'g':
    case 'm':
    case 'r':
    case 'w':
      return 1;
    case 'c':
    case 'h':
    case 'n':
    case 's':
    case 'x':
      return 2;
    case 'd':
    case 'j':
    case 'o':
    case 't':
    case 'y':
      return 3;
    case 'e':
    case 'k':
    case 'p':
    case 'u':
    case 'z':
      return 4;
  }
}


static int
letter_to_y (char c)
{
  switch (c | 32) {
    case 'a':
    case 'b':
    case 'c':
    case 'd':
    case 'e':
      return 4;
    case 'f':
    case 'g':
    case 'h':
    case 'j':
    case 'k':
      return 3;
    case 'l':
    case 'm':
    case 'n':
    case 'o':
    case 'p':
      return 2;
    case 'q':
    case 'r':
    case 's':
    case 't':
    case 'u':
      return 1;
    case 'v':
    case 'w':
    case 'x':
    case 'y':
    case 'z':
      return 0;
  }
}

static char
xy_to_letter (int x, int y)
{
  char
  tbl[5][5] = { {'V', 'W', 'X', 'Y', 'Z'},
                {'Q', 'R', 'S', 'T', 'U'},
                {'L', 'M', 'N', 'O', 'P'},
                {'F', 'G', 'H', 'J', 'K'},
                {'A', 'B', 'C', 'D', 'E'}
              };

  if (x < 0)
    return ' ';
  if (y < 0)
    return ' ';
  if (x > 4)
    return ' ';
  if (y > 4)
    return ' ';

  return tbl[y][x];
}


class BOSGRID
{
  static BOSGRID BOSGRID::bosxy_to_bosgrid (BOSXY xy);

public:
  char big_zone;
  char small_zone;
  double e;
  double n;

  BOSGRID () {};

  BOSGRID (BOSXY xy)
  {
    *this = bosxy_to_bosgrid (xy);
  }

};


static BOSXY bosgrid_to_bosxy (BOSGRID g)
{
  BOSXY r (g.e, g.n);


  r.x += (double) (100000 * letter_to_x (g.small_zone));
  r.y += (double) (100000 * letter_to_y (g.small_zone));

  r.x += (double) (500000 * letter_to_x (g.big_zone));
  r.y += (double) (500000 * letter_to_y (g.big_zone));


  r.x -= 1000000.0;
  r.y -= 500000.0;

  return r;
}


static BOSGRID bosxy_to_bosgrid (BOSXY xy)
{
  BOSGRID r;
  int x, y;

  xy.x += 1000000.0;
  xy.y += 500000.0;

  x = (int) (xy.x / 500000.0);
  y = (int) (xy.y / 500000.0);

  r.big_zone = xy_to_letter (x, y);

  xy.x -= (double) (x * 500000);
  xy.y -= (double) (y * 500000);

  x = (int) (xy.x / 100000);
  y = (int) (xy.y / 100000);

  r.small_zone = xy_to_letter (x, y);

  xy.x -= (double) (x * 100000);
  xy.y -= (double) (y * 100000);

  r.e = xy.x;
  r.n = xy.y;

  return r;
}



/* BOS coordinate system */

string BosCoordSystem::name (void)
{
  char buf[1024];
  sprintf (buf, "British Ordinance Survery grid references");

  return buf;
}

string BosCoordSystem::lltocs (LatLon ll)
{
  BOSGRID g = bosxy_to_bosgrid (wgs84_to_bosxy (lltowgs84 (ll)));
  char buf[1024];

  sprintf (buf, "%c%c%05d%05d", g.big_zone, g.small_zone, (int) (g.e + 0.5),
           (int) (g.n + 0.5));

  return buf;

}


LatLon BosCoordSystem::cstoll (string s)
{
  LatLon ll;
  BOSGRID g;
  int i = s.length ();
  int j;

  if (i % 2)
    return ll;

  g.big_zone = s[0];
  g.small_zone = s[1];

  i -= 2;
  i /= 2;

  g.e = g.n = 0;

  for (j = 0; j < i; ++j) {
    g.e *= 10.0;
    g.e += (double) s[j + 2] - '0';
    g.n *= 10.0;
    g.n += (double) s[i + j + 2] - '0';
  }
  for (; j < 5; ++j) {
    g.n *= 10.0;
    g.e *= 10.0;
  }

  WGS84 w = bosxy_to_wgs84 (bosgrid_to_bosxy (g));

  return LatLon (w.lat, w.lon);
}


/* BOSXY coordinate system */

string BosxyCoordSystem::name (void)
{
  char buf[1024];
  sprintf (buf, "British Ordinance Survery xy");

  return buf;
}

string BosxyCoordSystem::lltocs (LatLon ll)
{
  BOSXY xy = wgs84_to_bosxy (lltowgs84 (ll));
  char buf[1024];

  sprintf (buf, "%.1f,%.1f", xy.x, xy.y);

  return buf;

}


LatLon BosxyCoordSystem::cstoll (string s)
{
  char buf[1024], *ptr;
  BOSXY xy;

  strcpy (buf, s.c_str ());

  if (!(ptr = index (buf, ',')))
    return LatLon(0, 0);

  *ptr = 0;
  xy.x = atof(buf);
  xy.y = atof(ptr + 1);

  WGS84 w = bosxy_to_wgs84 (xy);

  return wgs84toll(w);
}


/* Map object */

class BosTileDescription: public TileDescription
{
public:
  BosTileDescription (BOSXY ixy, int iw, int ih)
  {
    xy = ixy;
    w = iw;
    h = ih;
  }
  int issame (TileDescription * it1, TileDescription * it2)
  {
    BosTileDescription * t1 = (BosTileDescription *) it1;
    BosTileDescription * t2 = (BosTileDescription *) it2;

    if (t1->xy.x != t2->xy.x)
      return 0;
    if (t1->xy.y != t2->xy.y)
      return 0;
    if (t1->w != t2->w)
      return 0;
    if (t1->h != t2->h)
      return 0;
    return 1;
  };
  BOSXY xy;
  int w;
  int h;
};

class BosMap: public Map
{
public:
  string name (void);
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort);
  ~BosMap () {};

  double scale;
  BosImgCoordSystem bics;

  TileCache tc;

  virtual Tile * load_tile (BosTileDescription * td, int effort) = 0;
};

class BosMap1k: public BosMap
{
public:
  BosMap1k ()
  {
    scale = 1000;
    preferedics = new BosImgCoordSystem;
    preferedscale = 5;
    tiled = 1;
  };
  Tile * load_tile (BosTileDescription * td, int effort)
  {
    if ((effort == RENDER_LAZY) && ((td->w * td->h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      return NULL;
    }
    if ((td->w <= 0) || (td->h <= 0))
      return NULL;
    BOSGRID g = bosxy_to_bosgrid (td->xy);

    int be, bn;
    int se, sn;
    char mn, me;

    be = (int) (g.e / 10000.0);
    bn = (int) (g.n / 10000.0);

    g.e -= (double) (be * 10000);
    g.n -= (double) (bn * 10000);

    if (g.e < 5000.0) {
      me = 'W';
    } else {
      me = 'E';
      g.e -= 5000.0;
    }


    if (g.n < 5000.0) {
      mn = 'S';
    } else {
      mn = 'N';
      g.n -= 5000.0;
    }

    se = (int) (g.e / 1000.0);
    sn = (int) (g.n / 1000.0);

    g.e -= (double) (se * 1000);
    g.n -= (double) (sn * 1000);

    se *= 2;
    sn *= 2;

    char fn[1024];
    sprintf (fn, "%s/1km/%c%c/%d%d/%c%c%d%d%c%c%d%d.gif",
             getenv ("BOSPATH"),
             g.big_zone,
             g.small_zone,
             be, bn, g.big_zone, g.small_zone, be, bn, mn, me, se, sn);

    //    printf("Loading from %s\n", fn);

    GdkPixbuf * b = gdk_pixbuf_new_from_file (fn);
    if (!b)
      return NULL;
    GdkPixbuf * sb = gdk_pixbuf_scale_simple (b, td->w, td->h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);

    if (!sb)
      return NULL;

    Tile * t = new Tile;
    t->td = new BosTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
  ~BosMap1k ()
  {
    delete
    preferedics;
  };
};


class BosMap10k: public BosMap
{
public:
  BosMap10k ()
  {
    scale = 10000;
    preferedics = new BosImgCoordSystem;
    preferedscale = 50.0;
    tiled = 1;
  };
  Tile * load_tile (BosTileDescription * td, int effort)
  {

    if ((effort == RENDER_LAZY) && ((td->w * td->h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      return NULL;
    }
    if ((td->w <= 0) || (td->h <= 0))
      return NULL;
    BOSGRID g = bosxy_to_bosgrid (td->xy);
    int be, bn;
    int se, sn;
    char mn, me;

    be = (int) (g.e / 10000.0);
    bn = (int) (g.n / 10000.0);

    g.e -= (double) (be * 10000);
    g.n -= (double) (bn * 10000);

    char fn[1024];
    sprintf (fn, "%s/10km/%c%c/%c%c%d%d.gif",
             getenv ("BOSPATH"),
             g.big_zone, g.small_zone, g.big_zone, g.small_zone, be, bn);

    //    printf("Loading from %s\n", fn);

    GdkPixbuf * b = gdk_pixbuf_new_from_file (fn);
    if (!b)
      return NULL;
    GdkPixbuf * sb = gdk_pixbuf_scale_simple (b, td->w, td->h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);

    if (!sb)
      return NULL;

    Tile *
    t = new Tile;
    t->td = new BosTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
  ~BosMap10k ()
  {
    delete
    preferedics;
  };
};


class BosMap20k: public BosMap
{
public:
  BosMap20k ()
  {
    scale = 20000;
    preferedics = new BosImgCoordSystem;
    preferedscale = 100;
    tiled = 1;
  };
  Tile *
  load_tile (BosTileDescription * td, int effort)
  {
    if ((effort == RENDER_LAZY) && ((td->w * td->h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      return NULL;
    }
    if ((td->w <= 0) || (td->h <= 0))
      return NULL;
    BOSGRID g = bosxy_to_bosgrid (td->xy);

    int be, bn;
    int se, sn;
    char mn, me;

    be = (int) (g.e / 20000.0);
    bn = (int) (g.n / 20000.0);

    g.e -= (double) (be * 20000);
    g.n -= (double) (bn * 20000);

    be *= 2;
    bn *= 2;

    char fn[1024];
    sprintf (fn, "%s/20km/%c%c/%c%c%d%d.gif",
             getenv ("BOSPATH"),
             g.big_zone, g.small_zone, g.big_zone, g.small_zone, be, bn);

    //    printf("Loading from %s\n", fn);

    GdkPixbuf * b = gdk_pixbuf_new_from_file (fn);

    if (!b)
      return NULL;

    GdkPixbuf * sb = gdk_pixbuf_scale_simple (b, td->w, td->h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);

    if (!sb)
      return NULL;

    Tile * t = new Tile;
    t->td = new BosTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
  ~BosMap20k ()
  {
    delete
    preferedics;
  };
};



class BosMap100k: public BosMap
{
public:
  BosMap100k ()
  {
    scale = 100000;
    preferedics = new BosImgCoordSystem;
    preferedscale = 200;
    tiled = 1;
  };
  Tile * load_tile (BosTileDescription * td, int effort)
  {
    if ((effort == RENDER_LAZY) && ((td->w * td->h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      return NULL;
    }
    if ((td->w <= 0) || (td->h <= 0))
      return NULL;
    BOSGRID g = bosxy_to_bosgrid (td->xy);

    char fn[1024];
    sprintf (fn, "%s/100km/%c%c.gif",
             getenv ("BOSPATH"), g.big_zone, g.small_zone);

    //    printf ("Loading from %s\n", fn);

    GdkPixbuf * b = gdk_pixbuf_new_from_file (fn);
    if (!b)
      return NULL;
    GdkPixbuf * sb = gdk_pixbuf_scale_simple (b, td->w, td->h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);

    if (!sb)
      return NULL;

    Tile * t = new Tile;
    t->td = new BosTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
  ~BosMap100k ()
  {
    delete
    preferedics;
  };
};



Map *
create_bosmap (int s)
{
  switch (s) {
    case 1000:
      return new BosMap1k;
      break;
    case 10000:
      return new BosMap10k;
      break;
    case 20000:
      return new BosMap20k;
      break;
    case 100000:
      return new BosMap100k;
      break;
    default:
      FATALERROR ("create_bos_map unknown scale");
      break;
  }
}

string BosMap::name (void)
{
  char buf[1024];
  sprintf (buf, "British Ordinance Survey Map scale %.fkm",
           (double) (scale / 1000.0));
  return buf;
}

#define BOUND(v,l,h) v= (((v)>(h)) ? (h):( ((v)<(l))?(l):v ))



void
BosMap::render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort)
{

  BOSXY ll = bictobosxy (bics.lltoic (vcs->vctoll (ViewCoord (p->x, p->y + p->h))));
  BOSXY ur = bictobosxy (bics.lltoic (vcs->vctoll (ViewCoord (p->x + p->w, p->y))));

  ll.x -= CORRECT_X;
  ur.x -= CORRECT_X;
  ll.y -= CORRECT_Y;
  ur.y -= CORRECT_Y;


  ll.x = round_down (ll.x, scale);
  ll.y = round_down (ll.y, scale);

  BOSXY xy;
  xy.y = (ur.y - ll.y) / scale;
  xy.x = (ur.x - ll.x) / scale;
  xy.y *= xy.x;

  if ((effort == RENDER_LAZY) && (xy.y > MAXSUBTILES)) {
    printf ("MAXSUBTILES stitching limit exceeded tile not done\n");
    return ;
  }


  for (xy.y = ll.y; xy.y < ur.y; xy.y += scale) {
    for (xy.x = ll.x; xy.x < ur.x; xy.x += scale) {


      ViewCoord vc_ul = vcs->lltovc (bics.ictoll (bosxytobic (BOSXY (xy.x + CORRECT_X, xy.y + scale + CORRECT_Y))));
      ViewCoord vc_lr = vcs->lltovc (bics.ictoll (bosxytobic (BOSXY (xy.x + scale + CORRECT_X, xy.y + CORRECT_Y))));
      int tw = (int) ((vc_lr.x - vc_ul.x) + .5);
      int th = (int) ((vc_lr.y - vc_ul.y) + .5);


      BosTileDescription td (xy, tw, th);

      Tile *t = tc.find (&td);
      if (!t)
        t = load_tile (&td, effort);

      if ((t) && (!t->b))
        FATALERROR ("Fish");

      double dsx = (int) (vc_ul.x + .5);
      double dsy = (int) (vc_ul.y + .5);

      BOUND (vc_ul.x, p->x, p->x + p->w);
      BOUND (vc_ul.y, p->y, p->y + p->h);
      BOUND (vc_lr.x, p->x, p->x + p->w);
      BOUND (vc_lr.y, p->y, p->y + p->h);

      int x = (int) (vc_ul.x + .5);
      int y = (int) (vc_ul.y + .5);

      int sx = (int) (0.5 + vc_ul.x - dsx);
      int sy = (int) (0.5 + vc_ul.y - dsy);

      int w = (int) ((vc_lr.x - vc_ul.x) + .5);
      int h = (int) ((vc_lr.y - vc_ul.y) + .5);

      if (w && h) {
        if (t) {
          if ((w + sx) > tw)
            w -= (w + sx) - tw;
          if ((h + sy) > th)
            h -= (h + sy) - th;

          d->render (t->b, sx, sy, x, y, w, h, x, y);
        } else {
          if ((w > 0) || (h > 0))
            d->zap (x, y, w, h);
        }
      }

    }
  }
}
