/*
* bsm.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] = "$Id: bsm.cc,v 1.4 2002/08/17 15:27:53 root Exp $";

/*
 * $Log: bsm.cc,v $
 * Revision 1.4  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.3  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.2  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.1  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.8  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.7  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.6  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.5  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.4  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.3  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.2  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.1  2001/12/06 04:26:39  root
 * #
 *
 */



#include "project.h"

#include "coords/libcoords.h"

#define TX 375
#define TY 186
#undef SHIFT
#define SX 51
#define SY 15

/* Utility functions*/

static ImageCoord
bosxytobic (BOSXY xy)
{
  return ImageCoord (xy.x, xy.y);
}

static BOSXY
bictobosxy (ImageCoord ic)
{
  return BOSXY (ic.x, ic.y);
}

static WGS84
lltowgs84 (LatLon ll)
{
  return WGS84 (ll.lat, ll.lon);
}

static LatLon
wgs84toll (WGS84 w)
{
  return LatLon (w.lat, w.lon);
}


/* Map object */

class BsmTileDescription: public TileDescription
{
public:
  BsmTileDescription (BOSXY ixy, int iw, int ih)
  {
    xy = ixy;
    w = iw;
    h = ih;
  }
  int issame (TileDescription * it1, TileDescription * it2)
  {
    BsmTileDescription *t1 = (BsmTileDescription *) it1;
    BsmTileDescription *t2 = (BsmTileDescription *) it2;

    if (t1->xy.x != t2->xy.x)
      return 0;
    if (t1->xy.y != t2->xy.y)
      return 0;
    if (t1->w != t2->w)
      return 0;
    if (t1->h != t2->h)
      return 0;
    return 1;
  };
  BOSXY xy;
  int w;
  int h;
};

class BsmMap: public Map
{
public:
  string name (void);
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort);

  BosImgCoordSystem bics;

  TileCache tc;
  float sclx, scly;
  float scale;
  char sclc[80];

  float corx, cory;
  int dirx, diry;


  BsmMap (float iscale)
  {
    preferedics = new BosImgCoordSystem;

    preferedscale = scale = iscale;
    tiled = 1;

#if 0

    if (scale == 0.25) {
      corx = -3.0;
      cory = -13.0;
      sclx = 100;
      scly = 50;
      dirx = 1000;
      diry = 1000;
      strcpy(sclc, "0.25");
      corx -= sclx;
    } else
#endif
      if (scale == 1.25) {
        corx = 65;
        cory = 39;
        sclx = 500;
        scly = 250;
        dirx = 1000;
        diry = 1000;
        strcpy(sclc, "1.25");
        corx -= sclx;
      }
#if 0
      else if (scale == 6.25) {
        //FIXME;
        corx = 0;
        cory = 0;
        sclx = 2500;
        scly = 1250;
        dirx = 1000;
        diry = 1000;
        strcpy(sclc, "6.25");
        corx -= sclx;
      } else if (scale == 25) {
        corx = 742;
        cory = 455;
        sclx = 10000;
        scly = 5000;
        dirx = 100000;
        diry = 100000;
        strcpy(sclc, "25");
        corx -= sclx;
      }
#endif
      else {
        FATALMSG("Bsmmap unknown scale");
      }

    corx += (((float) SX) / ((float) TX)) * sclx;
    cory -= (((float) SY) / ((float) TY)) * scly;


  };
  ~BsmMap ()
  {
    delete preferedics;
  }


  Tile *load_tile (BsmTileDescription * td, int effort)
  {
    int bx, by;
    int x, y;
    int w, h;

    x = ((int )sclx) * (int) ((td->xy.x / sclx) + .5);
    y = ((int) scly ) * (int) ((td->xy.y / scly) + .5);

    bx = x / dirx;
    by = y / diry;


    char fn[1024];
    sprintf (fn, "%s/%s/%d/%d/%d.%d.gif",
             getenv ("BSMPATH"), sclc, bx, by, x, y);


    GdkPixbuf *b = gdk_pixbuf_new_from_file (fn);
    if (!b)
      return NULL;

    w = ((td->w) * gdk_pixbuf_get_width (b)) / TX;
    h = ((td->h) * gdk_pixbuf_get_height (b)) / TY;

    if ((effort == RENDER_LAZY) && ((w * h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      gdk_pixbuf_finalize (b);
      return NULL;
    }
#ifdef SHIFT
    gdk_pixbuf_copy_area (b, SX, SY, gdk_pixbuf_get_width (b) - SX , gdk_pixbuf_get_height (b) - SY, b, 0, 0);
#endif

    GdkPixbuf *sb = gdk_pixbuf_scale_simple (b, w, h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);
    if (!sb)
      return NULL;

    Tile *t = new Tile;
    t->td = new BsmTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
};



Map *
create_bsmmap (float scale)
{
  return new BsmMap(scale);
}

string BsmMap::name (void)
{
  char buf[1024];

  if (scale < 1.00) {
    sprintf (buf, "MM %d cm street map", (int) ((scale*100) + 0.5));
  } else {
    sprintf (buf, "MM %.2f m street map", scale);
  }
  return buf;
}

#define BOUND(v,l,h) v= (((v)>(h)) ? (h):( ((v)<(l))?(l):v ))


void
BsmMap::render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort)
{

  BOSXY ll =
    bictobosxy (bics.lltoic (vcs->vctoll (ViewCoord (p->x, p->y + p->h))));
  BOSXY ur =
    bictobosxy (bics.lltoic (vcs->vctoll (ViewCoord (p->x + p->w, p->y))));

  ll.x -= corx;
  ur.x -= corx;
  ll.y -= cory;
  ur.y -= cory;

  ll.x = round_down (ll.x, sclx);
  ll.y = round_down (ll.y, scly);


  BOSXY xy;
  xy.y = (ur.y - ll.y) / sclx;
  xy.x = (ur.x - ll.x) / scly;
  xy.y *= xy.x;

  if ((effort == RENDER_LAZY) && (xy.y > MAXSUBTILES)) {
    printf ("MAXSUBTILES stitching limit exceeded tile not done (%f)\n",
            xy.y);
    return ;
  }

  for (xy.y = ll.y; xy.y < ur.y; xy.y += scly) {
    for (xy.x = ll.x; xy.x < ur.x; xy.x += sclx) {


      ViewCoord vc_ul =
        vcs->lltovc (bics.ictoll (bosxytobic (BOSXY (xy.x + corx,
                                              xy.y + cory +
                                              scly))));
      ViewCoord vc_lr =
        vcs->lltovc (bics.
                     ictoll (bosxytobic (BOSXY (xy.x + corx + sclx,
                                                xy.y + cory))));
      int tw = (int) ((vc_lr.x - vc_ul.x) + .5);
      int th = (int) ((vc_lr.y - vc_ul.y) + .5);


      BsmTileDescription td (xy, tw, th);

      Tile *t = tc.find (&td);
      if (!t)
        t = load_tile (&td, effort);


      double dsx = (int) (vc_ul.x + .5);
      double dsy = (int) (vc_ul.y + .5);

      BOUND (vc_ul.x, p->x, p->x + p->w);
      BOUND (vc_ul.y, p->y, p->y + p->h);
      BOUND (vc_lr.x, p->x, p->x + p->w);
      BOUND (vc_lr.y, p->y, p->y + p->h);

      int x = (int) (vc_ul.x + .5);
      int y = (int) (vc_ul.y + .5);

      int sx = (int) (0.5 + vc_ul.x - dsx);
      int sy = (int) (0.5 + vc_ul.y - dsy);

      int w = (int) ((vc_lr.x - vc_ul.x) + .5);
      int h = (int) ((vc_lr.y - vc_ul.y) + .5);

      if (w && h) {
        if (t) {
          if ((w + sx) > tw)
            w -= (w + sx) - tw;
          if ((h + sy) > th)
            h -= (h + sy) - th;

          d->render (t->b, sx, sy, x, y, w, h, x, y);
        } else {
          d->zap (x, y, w, h);
        }
      }

    }
  }
}
