/*
* cache.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] =
  "$Id: cache.cc,v 1.11 2002/08/15 11:41:06 root Exp $";

/*
 * $Log: cache.cc,v $
 * Revision 1.11  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.10  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.9  2001/12/16 17:56:08  root
 * #
 *
 * Revision 1.8  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.7  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.6  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.5  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.4  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.3  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.2  2001/12/06 04:26:39  root
 * #
 *
 * Revision 1.1  2001/12/06 01:02:21  root
 * #
 *
 */

#include "project.h"

Tile::~Tile ()
{
  if (prev) {
    prev->next = next;
  } else {
    parent->head = next;
  }

  if (next) {
    next->prev = prev;
  } else {
    parent->tail = prev;
  }

  parent->numtiles--;
  parent->numbytes -= size;

  delete td;
  gdk_pixbuf_finalize (b);
  //  printf ("Tile at %p deleted\n", this);

}


TileCache::TileCache ()
{
  serial = 0;
  head = 0;
  tail = 0;
  numtiles = 0;
  numbytes = 0;
  maxtiles = MAXTILES;
  maxbytes = MAXTILESBYTES;
}


void
TileCache::add (Tile * t)
{
  //FIXME: should probably check tile not in cache;
  Tile *oldest;
  int old;

  if (head) {
    head->prev = t;
    t->next = head;
    t->next->prev = t;
    t->prev = 0;
    head = t;
  } else {
    head = t;
    tail = t;
    t->next = 0;
    t->prev = 0;
  }

  t->parent = this;
  t->lastuse = serial++;

  //printf("Tile at %p added\n",t);

  numtiles++;
  numbytes += t->size;


  while ((numtiles > maxtiles) || (numbytes > maxbytes)) {
    Tile *ct = head;

    old = serial;
    oldest = 0;

    while (ct) {

      if (ct->lastuse < old) {
        old = ct->lastuse;
        oldest = ct;
      }
      ct = ct->next;
    }

    /* One tile may be bigger than the cache */
    if (ct == t)
      return ;

    if (oldest)
      delete oldest;
  }
}

Tile *
TileCache::find (TileDescription * td)
{
  Tile *tp = head;


  while (tp) {
    if (tp->td->issame (tp->td, td)) {
      tp->lastuse = serial++;
      return tp;
    }
    tp = tp->next;
  }
  return 0;
}

TileCache::~TileCache ()
{
  while (head) {
    delete head;
  }
}
