
#define CORRECT_X 0.0
#define CORRECT_Y 0.0

#include "project.h"

#include "coords/libcoords.h"


/* Utility functions*/

static ImageCoord
bosxytobic (BOSXY xy)
{
  return ImageCoord (xy.x, xy.y);
}

static BOSXY
bictobosxy (ImageCoord ic)
{
  return BOSXY (ic.x, ic.y);
}

static WGS84
lltowgs84 (LatLon ll)
{
  return WGS84 (ll.lat, ll.lon);
}

static LatLon
wgs84toll (WGS84 w)
{
  return LatLon (w.lat, w.lon);
}


/* Map object */

class CazTileDescription: public TileDescription
{
public:
  CazTileDescription (BOSXY ixy, int iw, int ih)
  {
    xy = ixy;
    w = iw;
    h = ih;
  }
  int issame (TileDescription * it1, TileDescription * it2)
  {
    CazTileDescription *t1 = (CazTileDescription *) it1;
    CazTileDescription *t2 = (CazTileDescription *) it2;

    if (t1->xy.x != t2->xy.x)
      return 0;
    if (t1->xy.y != t2->xy.y)
      return 0;
    if (t1->w != t2->w)
      return 0;
    if (t1->h != t2->h)
      return 0;
    return 1;
  };
  BOSXY xy;
  int w;
  int h;
};

class CazMap: public Map
{
public:
  string name (void);
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort);

  BosImgCoordSystem bics;

  TileCache tc;

  CazMap ()
  {
    preferedics = new BosImgCoordSystem;
    preferedscale = 2.5;
    tiled = 1;
  };
  ~CazMap ()
  {
    delete preferedics;
  }


  Tile *load_tile (CazTileDescription * td, int effort)
  {
    int bx, by;
    int x, y;
    int w, h;
    BOSXY xy(td->xy);

    xy.x -= 540500;
    xy.y = 264000 - xy.y;
    x = (int) ((xy.x / 500) + .5);
    y = (int) ((xy.y / 500) + .5);


    //printf("(%f,%f) -> (%d,%d)\n", td->xy.x, td->xy.y, x, y);


    char fn[1024];
    sprintf (fn, "%s/%02d-%02d.jpg",
             getenv ("CAMAZPATH"), x, y);

    //    printf("Loading from %s\n", fn);


    GdkPixbuf *b = gdk_pixbuf_new_from_file (fn);
    if (!b)
      return NULL;

    w = td->w;
    h = td->h;

    if ((effort == RENDER_LAZY) && ((w * h) > MAXTILEPOINTS)) {
      printf ("MAXTILEPOINTS enlargement limit exceeded - tile not done\n");
      gdk_pixbuf_finalize (b);
      return NULL;
    }

    GdkPixbuf *sb = gdk_pixbuf_scale_simple (b, w, h, GDK_INTERP_BILINEAR);
    gdk_pixbuf_finalize (b);

    Tile *t = new Tile;
    t->td = new CazTileDescription (*td);
    t->b = sb;
    t->size = td->w * td->h * 3;

    tc.add (t);

    return t;
  };
};



Map *
create_cambridge_az_map (void)
{
  return new CazMap;
}

string CazMap::name (void)
{
  char
  buf[1024];
  sprintf (buf, "Cambridge AZ");
  return buf;
}

#define BOUND(v,l,h) v= (((v)>(h)) ? (h):( ((v)<(l))?(l):v ))


void
CazMap::render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort)
{

  BOSXY ll =
    bictobosxy (bics.lltoic (vcs->vctoll (ViewCoord (p->x, p->y + p->h))));
  BOSXY ur =
    bictobosxy (bics.lltoic (vcs->vctoll (ViewCoord (p->x + p->w, p->y))));

  ll.x -= CORRECT_X;
  ur.x -= CORRECT_X;
  ll.y -= CORRECT_Y;
  ur.y -= CORRECT_Y;

  ll.x = round_down (ll.x, 500);
  ll.y = round_down (ll.y, 500);


  BOSXY xy;
  xy.y = (ur.y - ll.y) / 500.0;
  xy.x = (ur.x - ll.x) / 500.0;
  xy.y *= xy.x;

  if ((effort == RENDER_LAZY) && (xy.y > MAXSUBTILES)) {
    printf ("MAXSUBTILES stitching limit exceeded tile not done (%f)\n",
            xy.y);
    return ;
  }

  for (xy.y = ll.y; xy.y < ur.y; xy.y += 500) {
    for (xy.x = ll.x; xy.x < ur.x; xy.x += 500) {


      ViewCoord vc_ul =
        vcs->lltovc (bics.ictoll (bosxytobic (BOSXY (xy.x + CORRECT_X,
                                              xy.y + CORRECT_Y +
                                              500))));
      ViewCoord vc_lr =
        vcs->lltovc (bics.
                     ictoll (bosxytobic (BOSXY (xy.x + CORRECT_X + 500,
                                                xy.y + CORRECT_Y))));
      int tw = (int) ((vc_lr.x - vc_ul.x) + .5);
      int th = (int) ((vc_lr.y - vc_ul.y) + .5);


      CazTileDescription td (xy, tw, th);

      Tile *t = tc.find (&td);
      if (!t)
        t = load_tile (&td, effort);


      double dsx = (int) (vc_ul.x + .5);
      double dsy = (int) (vc_ul.y + .5);

      BOUND (vc_ul.x, p->x, p->x + p->w);
      BOUND (vc_ul.y, p->y, p->y + p->h);
      BOUND (vc_lr.x, p->x, p->x + p->w);
      BOUND (vc_lr.y, p->y, p->y + p->h);

      int x = (int) (vc_ul.x + .5);
      int y = (int) (vc_ul.y + .5);

      int sx = (int) (0.5 + vc_ul.x - dsx);
      int sy = (int) (0.5 + vc_ul.y - dsy);

      int w = (int) ((vc_lr.x - vc_ul.x) + .5);
      int h = (int) ((vc_lr.y - vc_ul.y) + .5);



      if (w && h) {
        if (t) {
          if ((w + sx) > tw)
            w -= (w + sx) - tw;
          if ((h + sy) > th)
            h -= (h + sy) - th;

          d->render (t->b, sx, sy, x, y, w, h, x, y);
        } else {
          d->zap (x, y, w, h);
        }
      }

    }
  }
}
