/*
* cmd_p.cc:
*
* Copyright (c) 2002 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] = "$Id: cmd_p.cc,v 1.2 2002/08/17 15:27:53 root Exp $";

/*
 * $Log: cmd_p.cc,v $
 * Revision 1.2  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.1  2002/08/17 11:30:19  root
 * #
 *
 */

#include "project.h"

extern Map *maps[];
extern int nmaps;
extern CoordSystem *cs[];
extern int ncs;


extern void cmd_p_error(char *);
extern int cmd_p_parse(void);
extern int cmd_p_line;

char *cmd_p_name;
static FILE *cmd_p_file;
static char *filename = 0;

static CmdView *cv = 0;

static void check_cv(void)
{
  if (!cv)
    cv = new CmdView(maps, nmaps);
}

static void clean(void)
{
  if (cv) {
    delete cv;
    cv = 0;
  }
  if (filename) {
    free(filename);
    filename = 0;
  }
}
int cmd_p_read(char *buf, int len)
{
  return fread(buf, 1, len, cmd_p_file);
}

void cmd_p_addmap(char *map)
{
  check_cv();
  if (!cv->showmaps(map))
    cmd_p_error("No maps matched");
}

void cmd_p_removemap(char *map)
{
  check_cv();
  if (!cv->hidemaps(map))
    cmd_p_error("No maps matched");
}


void cmd_p_setpos(char *csn, char *where)
{
  int i, j = 0;
  check_cv();

  for (i = 0;i < ncs;++i) {
    if (strstr(cs[i]->name().c_str(), csn)) {
      j++;
      cv->move(cs[i]->cstoll(where));
      break; //Or better regexps
    }
  }

  if (!j)
    cmd_p_error("No coordinate systems matched");
  if (j != 1)
    cmd_p_error("More than one coordinate system matched");
}

void cmd_p_setscale(double s)
{
  check_cv();
  printf("Setting scale to %f\n", s);
  cv->setscale(s);
}

void cmd_p_setcs(char *cs)
{
  check_cv();
  if (!cv->setcs(cs)) {
    cmd_p_error("No coordinate systems matched");
  }
}

void cmd_p_setsize(int w, int h)
{
  check_cv();
  cv->setsize(w, h);
}

void cmd_p_setfn(char *fn)
{
  if (filename)
    free(filename);
  filename = strdup(fn);
}


void cmd_p_endbrace(void)
{
  check_cv();
  cv->render(filename);
  clean();
}

int cmd_parse(char *fn)
{
  clean();

  cmd_p_name = fn;
  cmd_p_file = fopen(fn, "r");
  if (!cmd_p_file)
    return -1;

  cmd_p_line = 0;

  cmd_p_parse();

  fclose(cmd_p_file);

  return 0;
}
