/*
* cmdview.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] =
  "$Id: cmdview.cc,v 1.6 2002/08/17 15:50:56 root Exp root $";

/*
 * $Log: cmdview.cc,v $
 * Revision 1.6  2002/08/17 15:50:56  root
 * #
 *
 * Revision 1.5  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.4  2002/08/17 11:30:19  root
 * #
 *
 * Revision 1.3  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.2  2002/08/15 11:41:18  root
 * #
 *
 * Revision 1.1  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.12  2002/08/12 15:42:48  root
 * #
 *
 * Revision 1.11  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.10  2001/12/16 17:56:08  root
 * #
 *
 * Revision 1.9  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.8  2001/12/16 17:42:43  root
 * #
 *
 * Revision 1.7  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.6  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.5  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.4  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.3  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.2  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.1  2001/12/06 04:26:39  root
 * #
 *
 */

#include "project.h"


ViewCoordSystem CmdView::make_vcs(void)
{
  ViewCoordSystem vcs;
  ImageCoord ic;

  ic = imgcoordsys->lltoic (pos);

  vcs.ox = ic.x - ((scale * (double) w) / 2.0);
  vcs.oy = ic.y - ((scale * (double) h) / 2.0);
  vcs.sx = scale;
  vcs.sy = scale;
  vcs.h = (double) h;
  vcs.ics = imgcoordsys;

  vcs.cursor.pos = pos;
  vcs.cursor.heading = 0;

  return vcs;
}


GdkPixbuf *CmdView::do_render(int slice_x, int slice_y, int slice_w, int slice_h, int slice, int nslices)
{
  GdkPixbuf *pb;

  pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, slice_w /*+ TILESIZE*/, slice_h /*+ TILESIZE*/);
  double fw = (double) slice_w;
  double fh = (double) slice_h;

  GisDrawableGdkPB gd (pb, slice_w, slice_h, slice_x, slice_y, w, h);

  ViewCoordSystem vcs = make_vcs();

  vcs.shift(ViewCoord((double) slice_x, (double) slice_y));

  int i, m, n = 0;

  for (i = 0; i < nmaps; ++i) {
    if (usemaps[i] && maps[i]->tiled) {
      n++;
      m = i;
    }
  }

  if (!n) {
    return pb;
  }

  n--;


  GdkPixbuf *b1 = 0, *b2 = 0;
  if (n) {
    b1 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, TILESIZE, TILESIZE);
    b2 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, TILESIZE, TILESIZE);
  }
  GisDrawableGdkPB gdg1 (b1);
  GisDrawableGdkPB gdg2 (b2);

  double ntiles = 0, tiles = 0;

  for (double y = 0; y < fh; y += TILESIZE)
    for (double x = 0; x < fw; x += TILESIZE)
      ntiles++;

  ntiles = ntiles / 100.0;

  ImageCoord ic;
  for (double y = 0; y < fh; y += TILESIZE) {
    for (double x = 0; x < fw; x += TILESIZE) {
      char buf[80];
      ViewRect vr (x, y, TILESIZE, TILESIZE);

      printf("rendering: %.1f%% %d/%d\n", (tiles++) / ntiles, slice, nslices);

      if (n) {
        m = 0;
        for (i = 0; i < nmaps; ++i) {
          if (usemaps[i] && maps[i]->tiled) {
            if (!m) {
              gdg1.ox = (int) x;
              gdg1.oy = (int) y;
              maps[i]->render (&gdg1, &vcs, &vr, RENDER_TRY_HARDER);
              m = 1;
            } else {
              gdg2.ox = (int) x;
              gdg2.oy = (int) y;
              maps[i]->render (&gdg2, &vcs, &vr, RENDER_TRY_HARDER);

              gdk_pixbuf_composite (b2, b1, 0, 0, TILESIZE, TILESIZE, 0, 0, 1.0, 1.0, GDK_INTERP_NEAREST, 128);
            }
          }
        }

        gd.render (b1, 0, 0, (int) (x + .5), (int) (y + .5), (int) TILESIZE, (int) TILESIZE, (int) x, (int) y);

      } else {
        maps[m]->render (&gd, &vcs, &vr, RENDER_TRY_HARDER);
      }


    }
  }


  if (n) {
    gdk_pixbuf_finalize (b2);
    gdk_pixbuf_finalize (b1);
  }


  ViewRect vr (0, 0, w, h);

  for (i = 0; i < nmaps; ++i) {
    if (usemaps[i] && (!maps[i]->tiled)) {
      printf("Drawing %s\n", maps[i]->name().c_str());
      maps[i]->render(&gd, &vcs, &vr, RENDER_TRY_HARDER);
    }
  }



  return pb;

}

#if 0
#define SLICESIZE 50000
#else
#define SLICESIZE 64*1024*1024
#endif

int CmdView::render(FILE *f)
{
  GdkPixbuf *pb;
  long filesize = (w ) * (h ) * 3;
  int i, n;

  if (filesize < SLICESIZE) {
    pb = do_render(0, 0, w, h, 1, 1);
    if (!pb)
      return -1;
    printf("Writing out image\n");
    save_pixbuf_to_stream(pb, w, h, f);
    gdk_pixbuf_finalize (pb);
    return 0;
  }

  int slice_y = 0;
  int slice_h = (SLICESIZE / (3 * (w + TILESIZE)));

  slice_h &= ~0xf;

  if (!slice_h) {
	printf("bitmap to wide\n");
    return -1;
  }

  n = (h + slice_h - 1) / slice_h;
  i = 1;

  save_pixbuf_to_stream_incr_start(w, h, f);

  while (slice_y != h) {

    if ((slice_y + slice_h) > h)
      slice_h = h - slice_y;

    pb = do_render(0, slice_y, w, slice_h, i, n);

    if (!pb)
      return -1;

    printf("Writing out %d/%d\n", i++, n);
    save_pixbuf_to_stream_incr_save(pb, w, slice_h, f);
    gdk_pixbuf_finalize (pb);

    slice_y += slice_h;
  }

  printf("Done\n");

  return 0;
}

int CmdView::render(char *fn)
{
  FILE *f = fopen(fn, "w");
  int ret;

  if (!f)
    return -1;

  ret = render(f);
  fclose(f);

  return ret;
}


CmdView::~CmdView ()
{
  delete[] usemaps;
}

void CmdView::move(LatLon where)
{
  pos = where;
}

void CmdView::setscale(double is)
{
  scale = is;
}

void CmdView::setsize(int iw, int ih)
{
  w = iw;
  h = ih;
}

int CmdView::setcs(char *str)
{
  int i, j = 0;
  for (i = 0;i < nmaps;++i) {
    if (strstr(maps[i]->name().c_str(), str)) {
      imgcoordsys = maps[i]->preferedics;

      printf("Set Coordinate system to %s\n",
             imgcoordsys->name().c_str());
      j++;

    }
  }
  return j;
}

int CmdView::showmaps(char *str)
{
  int i, j = 0;
  for (i = 0;i < nmaps;++i) {
    if (strstr(maps[i]->name().c_str(), str)) {
      printf("Added %s\n", maps[i]->name().c_str());
      usemaps[i] = 1;
      j++;

    }
  }

  return j;
}

int CmdView::hidemaps(char *str)
{
  int i, j = 0;
  for (i = 0;i < nmaps;++i) {
    if (strstr(maps[i]->name().c_str(), str)) {
      printf("Removed %s\n", maps[i]->name().c_str());
      usemaps[i] = 0;
      j++;

    }
  }
  return j;
}


CmdView::CmdView (Map **mps, int n)
{
  maps = mps;
  nmaps = n;

  usemaps = new int[nmaps];
  int i;
  for (i = 0;i < nmaps;++i)
    usemaps[i] = 0;

  usemaps[0] = 1;
  usemaps[1] = 1;
  usemaps[2] = 1;


}

CmdView::CmdView (ImageCoordSystem *ics, Map **mps, int *which, int n, LatLon where, double iscale, int iw, int ih)
{
  pos = where;
  imgcoordsys = ics;
  scale = iscale;
  nmaps = n;
  maps = mps;

  usemaps = which;

  if (!which) {
    int i;
    usemaps = new int[nmaps];
    for (i = 0;i < nmaps;++i)
      usemaps[i] = 0;
  }


  w = iw;
  h = ih;


  //string s = imgcoordsys->name ();
  //printf("Created CmdView on Coordinate system %s\n", s.c_str());


  //  s = maps[i]->name ();
  //  s.c_str();

}
