/*
* color.cc:
*
* Copyright (c) 2002 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] = "$Id: color.cc,v 1.5 2002/08/17 15:50:56 root Exp root $";

/*
 * $Log: color.cc,v $
 * Revision 1.5  2002/08/17 15:50:56  root
 * #
 *
 * Revision 1.4  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.3  2002/08/15 11:41:18  root
 * #
 *
 * Revision 1.2  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.1  2002/08/12 15:42:48  root
 * #
 *
 */

#include "project.h"


void GisColor::itof(void)
{
  r = ir;
  r = r / 255.0;
  g = ig;
  g = g / 255.0;
  b = ib;
  b = b / 255.0;
}

void GisColor::ftoi(void)
{
  ir = (int) ((r * 255.0) + 0.5);
  ig = (int) ((g * 255.0) + 0.5);
  ib = (int) ((b * 255.0) + 0.5);

  if (ir < 0)
    ir = 0;
  if (ir > 255)
    ir = 255;
  if (ig < 0)
    ig = 0;
  if (ig > 255)
    ig = 255;
  if (ib < 0)
    ib = 0;
  if (ib > 255)
    ib = 255;

}

art_u32 GisColor::art(int alpha)
{
  return (((art_u32) ib) << 24) | (((art_u32) ig) << 16 ) | (((art_u32) ir) << 8) | (art_u32) alpha;
}

u_int32_t GisColor::pb(int alpha)
{
  u_int32_t ret;
  unsigned char *ptr = (unsigned char *) & ret;

  ptr[0] = ib;
  ptr[1] = ig;
  ptr[2] = ir;
  ptr[3] = alpha;

  return ret;
}

unsigned long GisColor::bgr()
{
  return (((unsigned long) ib) << 16) | (((unsigned long) ig) << 8 ) | (unsigned long) ir;
}

unsigned long GisColor::rgb()
{
  return (((unsigned long) ir) << 16) | (((unsigned long) ig) << 8 ) | (unsigned long) ib;
}

char * GisColor::ps()
{
  //FIXME:
  return strdup("");
}

void GisColor::brflip(void)
{
  double t;
  int it;

  t = r;
  r = b;
  b = t;

  it = ir;
  ir = ib;
  ib = it;

  return ;
}

guint32 GisColor::X(void)
{
  return bgr();
}


GisColor::GisColor()
{}

GisColor::GisColor(char *col)
{
  unsigned long i;
  if (col[0] == '#') {
    i = strtol(col + 1, NULL, 16);

    ib = (i >> 16) & 0xff;
    ig = (i >> 8) & 0xff;
    ir = i & 0xff;
    itof();

  } else {
    ir = ig = ib = 0;
    itof();

  }


}

GisColor::GisColor(int i)
{
  ib = (i >> 16) & 0xff;
  ig = (i >> 8) & 0xff;
  ir = i & 0xff;
  itof();
}


GisColor::GisColor(unsigned long i)
{
  ib = (i >> 16) & 0xff;
  ig = (i >> 8) & 0xff;
  ir = i & 0xff;
  itof();
}

GisColor::GisColor(int ur, int ug, int ub)
{
  ib = ur;
  ig = ug;
  ir = ub;
  itof();
}

GisColor::GisColor(float ur, float ug, float ub)
{
  r = (double) ur;
  g = (double) ug;
  b = (double) ub;
  ftoi();
}

GisColor::GisColor(double ur, double ug, double ub)
{
  r = ur;
  g = ug;
  b = ub;
  ftoi();
}

GisColor Cblack ( 0, 0, 0);
GisColor Cblue ( 0, 0, 255);
GisColor Cgreen ( 0, 255, 0);
GisColor Ccyan ( 0, 255, 255);
GisColor Cred (255, 0, 0);
GisColor Cmagenta(255, 0, 255);
GisColor Cyellow (255, 255, 0);
GisColor Cwhite (255, 255, 255);
GisColor Cgrey (128, 128, 128);
GisColor Cscale (0x10, 0x7c, 0xff);

