/*
* coordinates.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] =
  "$Id: coordinates.cc,v 1.10 2002/08/17 11:30:19 root Exp $";

/*
 * $Log: coordinates.cc,v $
 * Revision 1.10  2002/08/17 11:30:19  root
 * #
 *
 * Revision 1.9  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.8  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.7  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.6  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.5  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.4  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.3  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.2  2001/12/06 04:26:39  root
 * #
 *
 * Revision 1.1  2001/12/06 01:02:21  root
 * #
 *
 */

#include "project.h"

void ViewCoordSystem::shift(ViewCoord vc)
{
  ox += (vc.x * sx) ;
  oy -= (vc.y * sy);
}

ImageCoord ViewCoordSystem::vctoic (ViewCoord vc)
{
  ImageCoord
  ic;

  ic.x = (vc.x * sx) + ox;
  ic.y = ((h - vc.y) * sy) + oy;

  return ic;
}

ViewCoord ViewCoordSystem::ictovc (ImageCoord ic)
{
  ViewCoord
  vc;

  vc.x = (ic.x - ox) / sx;
  vc.y = h - ((ic.y - oy) / sy);

  return vc;
}

LatLon ViewCoordSystem::vctoll (ViewCoord vc)
{
  return ics->ictoll (vctoic (vc));
}

ViewCoord ViewCoordSystem::lltovc (LatLon ll)
{
  return ictovc (ics->lltoic (ll));
}

string
LatLonCoordSystem::name (void)
{
  char buf[1024];
  sprintf (buf, "WGS84 Lat/Lon");
  return string (buf);
}

string
LatLonCoordSystem::lltocs (LatLon ll)
{
  char buf[1024];
  sprintf (buf, "%.6f,%.6f", ll.lat, ll.lon);
  return string (buf);
}

LatLon
LatLonCoordSystem::cstoll (string s)
{
  char buf[1024], *ptr;
  LatLon ll (0, 0);
  strcpy (buf, s.c_str ());

  if (!(ptr = index (buf, ',')))
    return ll;
  *ptr = 0;
  ll.lat = atof (buf);
  ll.lon = atof (ptr + 1);
  return ll;
}

string
DMLatLonCoordSystem::name (void)
{
  char buf[1024];
  sprintf (buf, "DM WGS84 Lat/Lon");
  return string (buf);
}

string
DMLatLonCoordSystem::lltocs (LatLon ll)
{
  char buf[1024];
  int ad, od;
  int sa = 0, so = 0;


  if (ll.lat < 0) {
    sa++;
    ll.lat = -ll.lat;
  }
  if (ll.lon < 0) {
    so++;
    ll.lon = -ll.lon;
  }

  ad = (int) ll.lat;
  ll.lat -= (double) ad;
  ll.lat *= 60;

  od = (int) ll.lon;
  ll.lon -= (double) od;
  ll.lon *= 60;

  sprintf (buf, "%s%d'%.4f,%s%d'%.4f", sa ? "-" : "", ad, ll.lat, so ? "-" : "", od, ll.lon);
  return string (buf);
}


#define IABS(a)  (((a)>0) ? (a):-(a))

LatLon
DMLatLonCoordSystem::cstoll (string s)
{
  char buf[1024];
  LatLon ll (0, 0);
  float am, om;
  char sa[20], so[20];
  strcpy (buf, s.c_str ());

  sscanf (buf, "%[-0123456789]%*['\"]%f%*[,]%[-0123456789]%*['\"]%f", sa, &am, so, &om);

  ll.lat = am;
  ll.lat = ll.lat / 60.0;
  ll.lat += IABS(atoi(sa));

  if (sa[0] == '-')
    ll.lat = -ll.lat;

  ll.lon = om;
  ll.lon = ll.lon / 60.0;
  ll.lon += IABS(atoi(so));
  if (so[0] == '-')
    ll.lon = -ll.lon;

  return ll;
}


string
DMSLatLonCoordSystem::name (void)
{
  char buf[1024];
  sprintf (buf, "DMS WGS84 Lat/Lon");
  return string (buf);
}

string
DMSLatLonCoordSystem::lltocs (LatLon ll)
{
  char buf[1024];
  int ad, am, od, om;
  int sa = 0, so = 0;


  if (ll.lat < 0) {
    sa++;
    ll.lat = -ll.lat;
  }
  if (ll.lon < 0) {
    so++;
    ll.lon = -ll.lon;
  }

  ad = (int) ll.lat;
  ll.lat -= (double) ad;
  ll.lat *= 60;
  am = (int) ll.lat;
  ll.lat -= (double) am;
  ll.lat *= 60;

  od = (int) ll.lon;
  ll.lon -= (double) od;
  ll.lon *= 60;
  om = (int) ll.lon;
  ll.lon -= (double) om;
  ll.lon *= 60;

  sprintf (buf, "%s%d'%d\"%.2f,%s%d'%d\"%.2f", sa ? "-" : "", ad, am, ll.lat, so ? "-" : "", od, om, ll.lon);
  return string (buf);
}

LatLon
DMSLatLonCoordSystem::cstoll (string s)
{
  char buf[1024];
  LatLon ll (0, 0);
  int am, om;
  float as, os;
  char so[20], sa[20];
  strcpy (buf, s.c_str ());

  sscanf (buf, "%[-0123456789]%*['\"]%d%*['\"]%f%*[,]%[-0123456789]%*['\"]%d%*['\"]%f",
          sa, &am, &as, so, &om, &os);

  ll.lat = as;
  ll.lat = ll.lat / 60.0;
  ll.lat += am;
  ll.lat = ll.lat / 60.0;
  ll.lat += IABS(atoi(sa));

  if (sa[0] == '-')
    ll.lat = -ll.lat;

  ll.lon = os;
  ll.lon = ll.lon / 60.0;
  ll.lon += om;
  ll.lon = ll.lon / 60.0;
  ll.lon += IABS(atoi(so));

  if (so[0] == '-')
    ll.lon = -ll.lon;

  return ll;
}

double
d2r (double d)
{
  d *= M_PI / 180.0;
  return d;
}

double
r2d (double d)
{
  d *= 180.0 / M_PI;
  return d;
}



// XYZ stuff...
//FIXME: these need to do all the datum stuff properly and
//be put in libcoord
XYZ
lltoxyz (LatLon ll)
{
  XYZ r;
  double q = cos (d2r (ll.lat));

  r.x = q * cos (d2r (ll.lon));
  r.y = q * sin (d2r (ll.lon));
  r.z = sin (d2r (ll.lat));
  return r;
}

LatLon
xyztoll (XYZ xyz)
{
  LatLon r;
  double q = sqrt ((xyz.x * xyz.x) + (xyz.y * xyz.y));
  r.lat = r2d (atan2 (xyz.z, q));
  r.lon = r2d (atan2 (xyz.y, xyz.x));
  return r;
}

double
xyzdot (XYZ x1, XYZ x2)
{
  return (x1.x * x2.x) + (x1.y * x2.y) + (x1.z * x2.z);
}

XYZ
xyzcross (XYZ x1, XYZ x2)
{
  XYZ r;

  r.x = (x1.y * x2.z) - (x1.z * x2.y);
  r.y = (x1.z * x2.x) - (x1.x * x2.z);
  r.z = (x1.x * x2.y) - (x1.y * x2.x);

  return r;
}

XYZ
xyznorm (XYZ x)
{
  double q = sqrt ((x.x * x.x) + (x.y * x.y) + (x.z * x.z));
  x.x = x.x / q;
  x.y = x.y / q;
  x.z = x.z / q;
  return x;
}


// Find the distance between two points.

double
lldist (LatLon l1, LatLon l2)
{
  XYZ x1 = lltoxyz (l1);
  XYZ x2 = lltoxyz (l2);
  XYZ c1, c2;
  double d1, d2, a;

  c1 = xyznorm (xyzcross (x1, x2));
  c2 = xyznorm (xyzcross (x1, c1));

  d1 = xyzdot (x1, x2);
  d2 = xyzdot (c2, x2);

  a = atan2 (d2, d1) * RADIUS;
  if (a < 0)
    a = -a;

  return a;
}
