/*
* coordinates.h:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

/*
 * $Id: coordinates.h,v 1.15 2002/08/17 15:27:53 root Exp $
 */

/*
 * $Log: coordinates.h,v $
 * Revision 1.15  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.14  2002/08/17 11:30:19  root
 * #
 *
 * Revision 1.13  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.12  2001/12/16 17:56:08  root
 * #
 *
 * Revision 1.11  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.10  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.9  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.8  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.7  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.6  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.5  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.4  2001/12/06 04:26:39  root
 * #
 *
 * Revision 1.3  2001/12/06 01:02:21  root
 * #
 *
 * Revision 1.2  2001/12/05 20:13:37  root
 * #
 *
 * Revision 1.1  2001/12/05 17:59:56  root
 * #
 *
 */

#ifndef __COORDINATES_H__
#define __COORDINATES_H__

class LatLon
{
public:
  double lat;
  double lon;
  LatLon (double a, double b)
  {
    lat = a;
    lon = b;
  };
  LatLon () {} ;
};

class Place: public LatLon
{
public:
  string name;
  Place (double a, double b, string n)
  {
    lat = a;
    lon = b;
    name = n;
  };
};

class ImageCoord
{
public:
  double x;
  double y;
  ImageCoord (double a, double b)
  {
    x = a;
    y = b;
  };
  ImageCoord () {} ;
};

class ViewCoord
{
public:
  double x;
  double y;
  ViewCoord (double a, double b)
  {
    x = a;
    y = b;
  };
  ViewCoord () {} ;
};

class ViewRect
{
public:
  ViewRect (double a, double b, double c, double d)
  {
    x = a;
    y = b;
    w = c;
    h = d;
  };
  double x;
  double y;
  double w;
  double h;
};


class ImageCoordSystem
{
public:
  virtual string name (void) = 0;
  virtual ImageCoord lltoic (LatLon) = 0;
  virtual LatLon ictoll (ImageCoord) = 0;
  virtual ~ImageCoordSystem ()
  {}
  ;
};

class CoordSystem
{
public:
  virtual string name (void) = 0;
  virtual string lltocs (LatLon) = 0;
  virtual LatLon cstoll (string) = 0;
  virtual ~CoordSystem () {} ;
};

class ViewCoordSystem
{
public:
  ImageCoordSystem * ics;
  double ox, oy;
  double sx, sy;
  double h;
  void shift(ViewCoord vc);
  ImageCoord vctoic (ViewCoord vc);
  ViewCoord ictovc (ImageCoord ic);
  LatLon vctoll (ViewCoord vc);
  ViewCoord lltovc (LatLon ll);
  struct
  {
    LatLon pos;
    double heading;
  }
  cursor;
};


class LatLonCoordSystem: public CoordSystem
{
public:
  string name (void);
  string lltocs (LatLon);
  LatLon cstoll (string);
};


class DMLatLonCoordSystem: public CoordSystem
{
public:
  string name (void);
  string lltocs (LatLon);
  LatLon cstoll (string);
};




class DMSLatLonCoordSystem: public CoordSystem
{
public:
  string name (void);
  string lltocs (LatLon);
  LatLon cstoll (string);
};


class XYZ
{
public:
  XYZ () {} ;
  XYZ (double a, double b, double c)
  {
    x = a;
    y = b;
    z = c;
  };
  double x;
  double y;
  double z;
};

extern double d2r (double);
extern double r2d (double);

extern LatLon xyztoll (XYZ);
extern XYZ lltoxyz (LatLon);
extern double xyzdot (XYZ, XYZ);
extern XYZ xyzcross (XYZ, XYZ);


#define RADIUS 6370000.0
extern double lldist (LatLon, LatLon);



#endif /* __COORDINATES_H__ */
