/*
 * coords.h:
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 * $Id: coords.h,v 1.2 2001/12/05 16:39:22 root Exp $
 *
 */

#ifndef __COORDS_H_ /* include guard */
#define __COORDS_H_

/* struct datum:
 * Definition of a datum (name, ellipsoid parameters).
 */
struct datum {
    char *name;
    short ellipsoid;
    short dx;
    short dy;
    short dz;
};

/* struct ellipsoid:
 * Definition of an ellipsoid.
 */
struct ellipsoid {
    char *name;     /* name of ellipsoid */
    double a;       /* semi-major axis, meters */
    double invf;    /* 1/f */
};

/* Datums themselves are defined in datum.c. */

#define DATUM_WGS84     100
#define DATUM_IRISH     39
#define DATUM_OSGB      75

/* grid.cc */
extern int geog_to_osgb (double lat, double lon, double *x,   double *y,   int datum);
extern int osgb_to_geog (double x,   double y,   double *lat, double *lon, int datum);
extern int geog_to_irish(double lat, double lon, double *x,   double *y,   int datum);
extern int irish_to_geog(double x,   double y,   double *lat, double *lon, int datum);

/* latlong.cc */
extern void geog_to_wgs84(double *lat, double *lon, int datum);
extern void wgs84_to_geog(double *lat, double *lon, int datum);

/* tm.cc */
extern void geog_to_tm(double lat, double lon, double lat0, double lon0, double k0, double *x,   double *y,   int datum);
extern void tm_to_geog(double x,   double y,   double lat0, double lon0, double k0, double *lat, double *lon, int datum);

extern struct datum const gDatum[];
extern struct ellipsoid const gEllipsoid[];

#endif /* __COORDS_H_ */
