/*
 * main.c:
 * Simple conversion program for coordinates.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] =
  "$Id: libcoords.cc,v 1.12 2002/08/15 11:41:08 root Exp $";

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "libcoords.h"
#include "coords.h"

IOSXY
wgs84_to_iosxy (WGS84 w)
{
  IOSXY g (0, 0);
  double east, north;


  //WGS84 -> Airy
  wgs84_to_geog (&w.lat, &w.lon, DATUM_IRISH);



  /* Obtain OS grid reference. */
  if (!geog_to_irish (w.lat, w.lon, &g.x, &g.y, DATUM_WGS84))
    return g;


  return g;
}

BOSXY
wgs84_to_bosxy (WGS84 w)
{
  BOSXY g (0, 0);

  //WGS84 -> Airy
  wgs84_to_geog (&w.lat, &w.lon, DATUM_OSGB);

  /* Obtain OS grid reference. */
  if (!geog_to_osgb (w.lat, w.lon, &g.x, &g.y, DATUM_WGS84))
    return g;

  return g;
}

WGS84
bosxy_to_wgs84 (BOSXY g)
{
  WGS84 w (0, 0);

  /* OS -> AIRY */
  if (!osgb_to_geog (g.x, g.y, &w.lat, &w.lon, DATUM_WGS84))
    return w;

  //Airy -> WGS84
  geog_to_wgs84 (&w.lat, &w.lon, DATUM_OSGB);

  return w;
}

WGS84
iosxy_to_wgs84 (IOSXY g)
{
  WGS84 w (0, 0);

  /* IRISH->AIRY */
  if (!irish_to_geog (g.x, g.y, &w.lat, &w.lon, DATUM_WGS84))
    return w;


  //Airy -> WGS84
  geog_to_wgs84 (&w.lat, &w.lon, DATUM_IRISH);

  return w;
}
