#include "project.h"

#define LINEWIDTH 3
#define FONTSIZE 20
#define SCALEWIDTH 5

#define CS 30
#define CE 5

#define PXS ((CS+1)*2)
#define PYS ((CS+1)*2)
#define PXO ((CS+1))
#define PYO ((CS+1))


#if 0
void
draw_cursor (Display * dpy, Drawable win, GC gc, int x, int y, double heading)
{
  int ex, ey, sx, sy;
  float th = (M_PI / 180.f) * (float) heading;
  float fx, fy;



  fx = sin (th);
  fy = -cos (th);
  sx = (int) (((float) CS) * fx);
  ex = (int) (((float) CE) * fx);
  sy = (int) (((float) CS) * fy);
  ey = (int) (((float) CE) * fy);

  ex += x;
  ey += y;
  sx += x;
  sy += y;


  XSetForeground (dpy, gc, 0x0);
  {
    XGCValues gc_val;
    gc_val.line_width = 3;
    XChangeGC (dpy, gc, GCLineWidth, &gc_val);
  }


  //  XCopyArea(dpy, win, cursor_pixmap, gc, x - PXO, y - PYO, PXS, PYS, 0, 0);

  //FIXME get colors from colormap...

  XDrawLine (dpy, win, gc, x - CS, y, x - CE, y);
  XDrawLine (dpy, win, gc, x + CS, y, x + CE, y);
  XDrawLine (dpy, win, gc, x, y - CS, x, y - CE);
  XDrawLine (dpy, win, gc, x, y + CS, x, y + CE);

  XSetForeground (dpy, gc, 0x13ff);
  XDrawLine (dpy, win, gc, sx, sy, ex, ey);
  //FIXME get colors from colormap...
  XSetForeground (dpy, gc, 0x0);
}

void
draw_scale (Display * dpy, Drawable win, GC gc, double length, int l, int w,
            int h)
{
  int x = 10;
  int y = h - 10;
  char buf[1000];
  int i;

  //FIXME get colors from colormap...
  XSetForeground (dpy, gc, 0x13ff);

  {
    XGCValues gc_val;
    gc_val.line_width = 5;
    XChangeGC (dpy, gc, GCLineWidth, &gc_val);
  }

  XDrawLine (dpy, win, gc, x, y, x + l, y);

  if (length < 1000) {
    i = sprintf (buf, "%.fm", length);
  } else {
    i = sprintf (buf, "%.fkm", length / 1000.0);
  }



  XDrawString (dpy, win, gc, x, y - 5, buf, i);
  //FIXME get colors from colormap...
  XSetForeground (dpy, gc, 0x0);

}
#endif



static void
draw_cursor (GisDrawable *gd, ViewCoordSystem *vcs)
{
  int ex, ey, sx, sy;
  double th = (M_PI / 180.f) * vcs->cursor.heading;
  double fx, fy;

  ViewCoord vc;
  vc = vcs->lltovc (vcs->cursor.pos);

  int x = (int) (vc.x + .5);
  int y = (int) (vc.y + .5);

  fx = sin (th);
  fy = -cos (th);
  sx = (int) (((double) CS) * fx);
  ex = (int) (((double) CE) * fx);
  sy = (int) (((double) CS) * fy);
  ey = (int) (((double) CE) * fy);

  ex += x;
  ey += y;
  sx += x;
  sy += y;


  gd->line( x - CS, y, x - CE, y, Cblack, LINEWIDTH);
  gd->line( x + CS, y, x + CE, y, Cblack, LINEWIDTH);
  gd->line( x, y - CS, x, y - CE, Cblack, LINEWIDTH);
  gd->line( x, y + CS, x, y + CE, Cblack, LINEWIDTH);
  gd->line( sx, sy, ex, ey, Cscale, LINEWIDTH);
}

class CursorMap: public Map
{
public:
  string name(void)
  {
    return string("Cursor");
  }
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort)
  {
    draw_cursor (d, vcs);
  }
  CursorMap()
  {
    tiled = 0;
    preferedics = 0;
  };
};

Map * create_cursormap (void)
{
  return new CursorMap;
}

static void
draw_scale (GisDrawable *gd, ViewCoordSystem *vcs)
{
  double ps, psd;


  ps = lldist (vcs->vctoll (ViewCoord (0, 0)), vcs->vctoll (ViewCoord ((double) gd->w, 0)));

  psd = 1.0;

  while ((psd*10) < ps)
    psd *= 10;

  ps = (psd / ps) * (double) gd->w;



  int x = (0 - gd->sx) + 10;
  int y = (gd->rh - gd->sy) - 10;

  char buf[1000];
  int i;
  int l = (int) (ps + .5);

  gd->line (x, y, x + l, y, Cscale, SCALEWIDTH);

  if (psd < 1000) {
    i = sprintf (buf, "%.fm", psd);
  } else {
    i = sprintf (buf, "%.fkm", psd / 1000.0);
  }

  gd->text (buf, x, y - (2*SCALEWIDTH), Cscale, FONTSIZE);

}

class ScaleMap: public Map
{
public:
  string name(void)
  {
    return string("Scale bar");
  }
  void render (GisDrawable * d, ViewCoordSystem * vcs, ViewRect * p, int effort)
  {
    draw_scale (d, vcs);
  }
  ScaleMap()
  {
    tiled = 0;
    preferedics = 0;
  };
};

Map * create_scalemap (void)
{
  return new ScaleMap;
}

