/*
* drawable.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] =
  "$Id: drawable.cc,v 1.16 2002/08/17 15:50:56 root Exp root $";

/*
 * $Log: drawable.cc,v $
 * Revision 1.16  2002/08/17 15:50:56  root
 * #
 *
 * Revision 1.15  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.14  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.13  2002/08/15 11:41:18  root
 * #
 *
 * Revision 1.12  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.11  2002/08/12 15:42:48  root
 * #
 *
 * Revision 1.10  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.9  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.8  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.7  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.6  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.5  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.4  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.3  2001/12/06 04:26:39  root
 * #
 *
 * Revision 1.2  2001/12/06 01:02:21  root
 * #
 *
 * Revision 1.1  2001/12/05 20:13:37  root
 * #
 *
 */

#include "project.h"

#if 0
#define FONT 		times_ttf
#define FONT_LEN 	times_ttf_len
#define FONT 		comic_ttf
#define FONT_LEN 	comic_ttf_len
#else
#define FONT 		n021003l_pfb
#define FONT_LEN 	n021003l_pfb_len
#endif

extern "C"
{
# include "fonts.h"
# if 1
#   include <ft2build.h>
#   include FT_FREETYPE_H
#   include FT_GLYPH_H
# else
#   include <freetype/freetype.h>
# endif
};

class Freetype
{
public:
  FT_Library library;
  FT_Face face;
  Freetype()
  {
    FT_Init_FreeType(&library);
    FT_New_Memory_Face(library, FONT, FONT_LEN,0,&face);
  };
  ~Freetype()
  {
    FT_Done_FreeType(library);
  };
}
ft;

static void ft_bbox(char *s, int *w, int *h, int fs)
{

  *w = *h = 0;

  if (!s)
    return ;

  int num_glyphs = strlen(s);

  if (!num_glyphs)
    return ;

  int error;

  FT_Glyph glyph;
  FT_Vector pos;
  FT_UInt previous;
  FT_GlyphSlot slot = ft.face->glyph;  // a small shortcut
  FT_UInt glyph_index;
  FT_Bool use_kerning;

  int pen_x = 0;
  int pen_y = 0;
  int n;

  FT_BBox bbox;
  // initialise string bbox to "empty" values
  bbox.xMin = bbox.yMin = 32000;
  bbox.xMax = bbox.yMax = -32000;

  // for each glyph image, compute its bounding box, translate it,
  // and grow the string bbox

  error = FT_Set_Pixel_Sizes(
            ft.face,     /* handle to face object            */
            0,        /* pixel_width                      */
            fs );   /* pixel_height                     */

  previous = 0;
  use_kerning = FT_HAS_KERNING(ft.face);
  glyph_index = 0;

  for ( n = 0; n < num_glyphs; n++ ) {
    FT_BBox glyph_bbox;
    FT_UInt glyph_index = FT_Get_Char_Index( ft.face, s[n] );

    if ( use_kerning && previous && glyph_index ) {
      FT_Vector delta;

      FT_Get_Kerning( ft.face, previous, glyph_index,
                      ft_kerning_default, &delta );

      pen_x += delta.x >> 6;
    }

    previous = glyph_index;

    error = FT_Load_Glyph( ft.face, glyph_index, FT_LOAD_DEFAULT );
    error = FT_Get_Glyph( ft.face->glyph, &glyph );

    FT_Glyph_Get_CBox( glyph, ft_glyph_bbox_pixels, &glyph_bbox );


    glyph_bbox.xMin += pos.x;
    glyph_bbox.xMax += pos.x;
    glyph_bbox.yMin += pos.y;
    glyph_bbox.yMax += pos.y;

    if (glyph_bbox.xMin < bbox.xMin)
      bbox.xMin = glyph_bbox.xMin;

    if (glyph_bbox.yMin < bbox.yMin)
      bbox.yMin = glyph_bbox.yMin;

    if (glyph_bbox.xMax > bbox.xMax)
      bbox.xMax = glyph_bbox.xMax;

    if (glyph_bbox.yMax > bbox.yMax)
      bbox.yMax = glyph_bbox.yMax;

    FT_Done_Glyph(glyph);
    pen_x += slot->advance.x >> 6;
    pen_y += slot->advance.y >> 6;   // unuseful for now..
  }

  // check that we really grew the string bbox


  if ( bbox.xMin > bbox.xMax ) {
    bbox.xMin = 0;
    bbox.yMin = 0;
    bbox.xMax = 0;
    bbox.yMax = 0;
  }

  // return string bbox
#if 0
  *w = bbox.xMax - bbox.xMin;
  *h = bbox.yMax - bbox.yMin;
#else

  *w = bbox.xMax;
  *h = bbox.yMax;
#endif


}



static GdkPixbuf *ftbitmap_to_pixbuf(FT_Bitmap *bm, GisColor c)
{
  int bh = bm->rows;
  int bw = bm->width;
  if ((!bh) && (!bw)) return NULL;
  GdkPixbuf *pb = gdk_pixbuf_new(GDK_COLORSPACE_RGB, True, 8, bw, bh);
  int pr = gdk_pixbuf_get_rowstride(pb);
  int br = (bm->pitch > 0 ) ? bm->pitch : -bm->pitch;
  int ph = gdk_pixbuf_get_height(pb);
  unsigned char *pptr, *prptr, *bptr, *brptr;
  int n;

  {
    u_int32_t uc = c.pb(0xff);
    u_int32_t *uptr = (u_int32_t *) gdk_pixbuf_get_pixels(pb);

    n = (pr * ph) >> 2;

    while (n--)
      *(uptr++) = uc;
  }

  prptr = gdk_pixbuf_get_pixels(pb);
  prptr += 3;
  brptr = bm->buffer;

  while (bh--) {
    pptr = prptr;
    bptr = brptr;
    for (n = 0;n < bw;++n) {
      *pptr = *(bptr++);
      pptr += 4;
    }
    prptr += pr;
    brptr += br;
  }

  return pb;
}

static int  do_horrid_clipping(int sw,int sh,int &sx,int &sy,int dw,int dh,int &dx,int &dy,int &cw,int &ch)
{

if (sx<0) {
  dx-=sx;
  cw+=sx;
  sx=0;
}

if (sy<0) {
  dy-=sy;
  ch+=sy;
  sy=0;
}

if (dx<0) {
  sx-=dx;
  cw+=dx;
  dx=0;
}

if (dy<0) {
  sy-=dy;
  ch+=dy;
  dy=0;
}

if ((cw+sx)>sw) {
	cw=sw-sx;
}

if ((ch+sy)>sh) {
	ch=sh-sy;
}

if ((cw+dx)>dw) {
	cw=dw-dx;
}

if ((ch+dy)>dh) {
	ch=dh-dy;
}
	
if (cw && ch) return 0;

return 1;

}

/* 1 2 3 */
/* 4 0 6 */
/* 7 8 9 */


void ft_text(GisDrawable *d, char *s, int pen_x, int pen_y, GisColor c, int fs)
{
  if (!s)
    return ;

  int num_glyphs = strlen(s);

  if (!num_glyphs)
    return ;

  int error;

  FT_Glyph glyph;
  FT_Vector pos;
  FT_UInt previous;
  FT_GlyphSlot slot = ft.face->glyph;  // a small shortcut
  FT_UInt glyph_index;
  FT_Bool use_kerning;

  int n;



  error = FT_Set_Pixel_Sizes(
            ft.face,     /* handle to face object            */
            0,        /* pixel_width                      */
            fs );   /* pixel_height                     */



  previous = 0;
  use_kerning = FT_HAS_KERNING(ft.face);
  glyph_index = 0;

  for ( n = 0; n < num_glyphs; n++ ) {
    FT_UInt glyph_index = FT_Get_Char_Index( ft.face, s[n] );

    if ( use_kerning && previous && glyph_index ) {
      FT_Vector delta;

      FT_Get_Kerning( ft.face, previous, glyph_index,
                      ft_kerning_default, &delta );

      pen_x += delta.x >> 6;
    }
    previous = glyph_index;


    error = FT_Load_Glyph( ft.face, glyph_index, FT_LOAD_DEFAULT );
    error = FT_Get_Glyph( ft.face->glyph, &glyph );

    pos.x = pen_x;
    pos.y = pen_y;

    error = FT_Glyph_To_Bitmap( &glyph, ft_render_mode_normal, &pos, 0 );


    if (!error) {
      FT_BitmapGlyph bit = (FT_BitmapGlyph)glyph;
      FT_Bitmap *bm = &bit->bitmap;

      GdkPixbuf *cpb = ftbitmap_to_pixbuf(bm, c);
      int x = pen_x, y = pen_y;


      x += bit->left;
      y -= bit->top;

      if (cpb)  {
      d->composite(cpb, 0, 0, bm->width, bm->rows, x, y);

      gdk_pixbuf_finalize(cpb);
	}
    }



    FT_Done_Glyph(glyph);
    pen_x += slot->advance.x >> 6;
    pen_y += slot->advance.y >> 6;   // unuseful for now..
  }




}



int GisDrawable::inarea(int x, int y)
{
  if (y < 0) {
    if (x < 0)
      return 1;
    if (x > w)
      return 3;
    return 2;
  }

  if (y > h) {
    if (x < 0)
      return 7;
    if (x > w)
      return 9;
    return 8;
  }

  if (x < 0)
    return 4;
  if (x > w)
    return 6;
  return 0;
}

void GisDrawableX::SetColor(GisColor *c)
{
  xlib_rgb_gc_set_foreground(gc, c->X());
//  XSetForeground(dpy, gc, c->X());
}

GisDrawableX::GisDrawableX (Display * idpy, Window iw, GC igc, int uw, int uh)
{
  dpy = idpy;
  gc = igc;
  drw = iw;
  rw = w = uw;
  rh = h = uh;
  sx = sy = 0;

}

void
GisDrawableX::render (GdkPixbuf * b, int sx, int sy, int dx, int dy,
                      int rw, int rh, int dithx, int dithy)
{
  gdk_pixbuf_xlib_render_to_drawable (b, drw, gc, sx, sy, dx, dy, rw, rh,
                                      XLIB_RGB_DITHER_NORMAL, dithx, dithy);
}

void
GisDrawableX::zap (int dx, int dy, int zw, int zh)
{
  SetColor(&Cwhite);
  XFillRectangle (dpy, drw, gc, dx, dy, zw, zh);
}


void
GisDrawableX::line (int x1, int y1, int x2, int y2, GisColor c, int lw)
{
  SetColor(&c);
  {
    XGCValues gc_val;
    gc_val.line_width = lw;
    XChangeGC (dpy, gc, GCLineWidth, &gc_val);
  }
  XDrawLine(dpy, drw, gc, x1, y1, x2, y2);
}

void
GisDrawableX::text (char *s, int x, int y, GisColor c, int fs)
{
#if 0
  SetColor(&c);
  XDrawString(dpy, drw, gc, x, y, s, strlen(s));
#else
  ft_text(this, s, x, y, c, fs);
#endif
}

void
GisDrawableX::textsize (char *s, int fs, int *tw, int *th)
{
  ft_bbox(s, tw, th, fs);
}
void
GisDrawableX::composite(GdkPixbuf *s, int sx, int sy, int w, int h, int dx, int dy)
{

  gdk_pixbuf_xlib_render_to_drawable_alpha(s, drw, sx, sy, dx, dy, w, h, GDK_PIXBUF_ALPHA_FULL, 128, XLIB_RGB_DITHER_NORMAL, 0, 0);

}

GisDrawableGdkPB::GisDrawableGdkPB (GdkPixbuf * ib)
{
  b = ib;
  ox = oy = 0;
  if (b) {
    rw = w = gdk_pixbuf_get_width(b);
    rh = h = gdk_pixbuf_get_height(b);
    sx = sy = 0;
  }
}
GisDrawableGdkPB::GisDrawableGdkPB (GdkPixbuf * ib, int iw, int ih)
{
  b = ib;
  ox = oy = 0;
  rw = w = iw;
  rh = h = ih;
  sx = sy = 0;
}

GisDrawableGdkPB::GisDrawableGdkPB (GdkPixbuf * ib, int iw, int ih, int isx, int isy, int irw, int irh)
{
  b = ib;
  ox = oy = 0;
  w = iw;
  h = ih;
  rw = irw;
  rh = irh;
  sx = isx;
  sy = isy;
}
void
GisDrawableGdkPB::render (GdkPixbuf * sb, int sx, int sy, int dx,
                          int dy, int rw, int rh, int dithx, int dithy)
{

  dx-=ox;
  dy-=oy;

  if (do_horrid_clipping(gdk_pixbuf_get_width(sb),gdk_pixbuf_get_height(sb),sx,sy,w,h,dx,dy,rw,rh)) return;

  gdk_pixbuf_copy_area (sb, sx, sy, rw, rh, b, dx, dy);
}

void
GisDrawableGdkPB::zap (int dx, int dy, int zw, int zh)
{
  GdkPixbuf *sb;
  sb = gdk_pixbuf_new(GDK_COLORSPACE_RGB, False, 8, zw, zh);
  memset(gdk_pixbuf_get_pixels(sb), 255,
         gdk_pixbuf_get_height(sb)*gdk_pixbuf_get_rowstride(sb));


  dx-=ox;
  dy-=oy;

  if (do_horrid_clipping(gdk_pixbuf_get_width(sb),gdk_pixbuf_get_height(sb),sx,sy,w,h,dx,dy,zw,zh)) return;

  gdk_pixbuf_copy_area (sb, 0, 0, zw, zh, b, dx , dy );

  gdk_pixbuf_finalize(sb);

}

void
GisDrawableGdkPB::line (int x1, int y1, int x2, int y2, GisColor c, int lw)
{
  ArtVpath *vec = NULL;
  ArtSVP *svp = NULL;
  int rs = gdk_pixbuf_get_rowstride(b);
  unsigned char *buffer = (unsigned char *) gdk_pixbuf_get_pixels(b);

  vec = art_new (ArtVpath, 3);
  vec[0].code = ART_MOVETO;
  vec[0].x = x1;
  vec[0].y = y1;
  vec[1].code = ART_LINETO;
  vec[1].x = x2;
  vec[1].y = y2;
  vec[2].code = ART_END;

  svp = art_svp_vpath_stroke(vec, ART_PATH_STROKE_JOIN_MITER,
                             ART_PATH_STROKE_CAP_SQUARE,
                             lw,
                             2 * lw,
                             0.5);


  art_rgb_svp_alpha (svp, 0, 0, w, h,
                     c.art(0xff), buffer, rs,
                     NULL);


  art_svp_free(svp);
  art_free(vec);

}

void
GisDrawableGdkPB::text (char *s, int pen_x, int pen_y, GisColor c, int fs)
{
  ft_text(this, s, pen_x, pen_y, c, fs);
}

void
GisDrawableGdkPB::textsize (char *s, int fs, int *w, int *h)
{
  ft_bbox(s, w, h, fs);
}

void
GisDrawableGdkPB::composite(GdkPixbuf *src, int sx, int sy, int cw, int ch, int dx, int dy)
{


  if (do_horrid_clipping(gdk_pixbuf_get_width(src),gdk_pixbuf_get_height(src),sx,sy,w,h,dx,dy,cw,ch)) return;

  gdk_pixbuf_composite(src, b, dx, dy, cw, ch, dx-sx, dy-sy, 1.0, 1.0, GDK_INTERP_NEAREST, 255);
}

