/*
* view.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] =
  "$Id: view.cc,v 1.18 2002/08/17 15:50:56 root Exp root $";

/*
 * $Log: view.cc,v $
 * Revision 1.18  2002/08/17 15:50:56  root
 * #
 *
 * Revision 1.17  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.16  2002/08/17 11:30:19  root
 * #
 *
 * Revision 1.15  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.14  2002/08/15 11:41:18  root
 * #
 *
 * Revision 1.13  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.12  2002/08/12 15:42:48  root
 * #
 *
 * Revision 1.11  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.10  2001/12/16 17:56:08  root
 * #
 *
 * Revision 1.9  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.8  2001/12/16 17:42:43  root
 * #
 *
 * Revision 1.7  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.6  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.5  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.4  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.3  2001/12/06 17:47:08  root
 * #
 *
 * Revision 1.2  2001/12/06 16:18:04  root
 * #
 *
 * Revision 1.1  2001/12/06 04:26:39  root
 * #
 *
 */

#include "project.h"

#define VIEW_WIDTH  800
#define VIEW_HEIGHT 600
#define VIEW_PANEL_HEIGHT 70


static ViewCoordSystem vcs_from_view_and_canvas (View * v, Canvas c)
{
  ViewCoordSystem vcs;
  ImageCoord ic;
  double w = (double) (int) xv_get (c, WIN_WIDTH);
  double h = (double) (int) xv_get (c, WIN_HEIGHT);

  ic = v->imgcoordsys->lltoic (v->parent->pos);


  vcs.ox = ic.x - ((w * v->scale) / 2.0);
  vcs.oy = ic.y - ((h * v->scale) / 2.0);
  vcs.sx = v->scale;
  vcs.sy = v->scale;
  vcs.h = h;
  vcs.ics = v->imgcoordsys;

  return vcs;
}


static ViewCoordSystem vcs_for_print_from_view_canvas_and_pixbuf(View * v, Canvas c, GdkPixbuf *b)
{
  ViewCoordSystem vcs;
  ImageCoord ic;
  double cw = (double) (int) xv_get (c, WIN_WIDTH);
  double ch = (double) (int) xv_get (c, WIN_HEIGHT);
  double w = (double) gdk_pixbuf_get_width(b);
  double h = (double) gdk_pixbuf_get_height(b);
  double a = w / h;
  double ca = cw / ch;
  double scale;

  if (a > ca) {
    scale = h / ch;
  } else {
    scale = w / cw;
  }

  ic = v->imgcoordsys->lltoic (v->parent->pos);

  scale = v->scale / scale;

  vcs.ox = ic.x - ((w * scale) / 2.0);
  vcs.oy = ic.y - ((h * scale) / 2.0);
  vcs.sx = scale;
  vcs.sy = scale;
  vcs.h = h;
  vcs.ics = v->imgcoordsys;

  return vcs;
}



static void canvas_repaint (Canvas c)
{
  View *v = (View *) xv_get (c, XV_KEY_DATA, VIEW_KEY);
  v->repaint ();
}


void View::print(FILE *f, int iw, int ih)
{
  int *um = new int[parent->nmaps];
  int i;

  for (i = 0;i < parent->nmaps;++i)
    um[i] = usemaps[i];

  CmdView *cv = new CmdView(imgcoordsys, parent->maps, um, parent->nmaps, parent->pos, scale, iw, ih);

  cv->render(f);

  delete cv;

#if 0

  GdkPixbuf *pb;

  pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, iw + TILESIZE, ih + TILESIZE);
  double w = (double) iw;
  double h = (double) ih;
  double ps, psd;

  GisDrawableGdkPB gd (pb);

  ViewCoordSystem vcs = vcs_for_print_from_view_canvas_and_pixbuf(this, canvas, pb);

  psd = 1.0;

  while ((psd * 10) < ps)
    psd *= 10;
  ps = (psd / ps) * w;

  int i, m, n = 0;

  for (i = 0; i < parent->nmaps; ++i) {
    if (usemaps[i]) {
      n++;
      m = i;
    }
  }

  if (!n) {

    //XFillRectangle (dpy, win, gc, 0, 0, (int) w, (int) h); FIXME:
    save_pixbuf_to_stream(pb, iw, ih, f) ;

    gdk_pixbuf_finalize (pb);
    return ;
  }

  n--;


  GdkPixbuf *b1 = 0, *b2 = 0;
  if (n) {
    b1 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, TILESIZE, TILESIZE);
    b2 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, TILESIZE, TILESIZE);
  }
  GisDrawableGdkPB gdg1 (b1);
  GisDrawableGdkPB gdg2 (b2);

  double ntiles = 0, tiles = 0;

  for (double y = 0; y < h; y += TILESIZE)
    for (double x = 0; x < w; x += TILESIZE)
      ntiles++;

  ntiles = ntiles / 100.0;

  ImageCoord ic;
  for (double y = 0; y < h; y += TILESIZE) {
    for (double x = 0; x < w; x += TILESIZE) {
      char buf[80];
      ViewRect vr (x, y, TILESIZE, TILESIZE);

      sprintf(buf, "rendering: %.1f%%", (tiles++) / ntiles);
      xv_set(printstatus, PANEL_VALUE, buf, 0);
      notify_dispatch();

      if (n) {
        m = 0;
        for (i = 0; i < parent->nmaps; ++i) {
          if (usemaps[i]) {
            if (!m) {
              gdg1.ox = (int) x;
              gdg1.oy = (int) y;
              parent->maps[i]->render (&gdg1, &vcs, &vr);
              m = 1;
            } else {
              gdg2.ox = (int) x;
              gdg2.oy = (int) y;
              parent->maps[i]->render (&gdg2, &vcs, &vr);

              gdk_pixbuf_composite (b2, b1, 0, 0, TILESIZE, TILESIZE, 0, 0, 1.0, 1.0, GDK_INTERP_NEAREST, 128);
            }
          }
        }

        gd.render (b1, 0, 0, (int) (x + .5), (int) (y + .5), (int) TILESIZE, (int) TILESIZE, (int) x, (int) y);

      } else {
        parent->maps[m]->render (&gd, &vcs, &vr);
      }


    }
  }

  xv_set(printstatus, PANEL_VALUE, "spawning", 0);
  notify_dispatch();

  if (n) {
    gdk_pixbuf_finalize (b2);
    gdk_pixbuf_finalize (b1);
  }

  if (1)
    draw_markup(&gd, &vcs);

  ViewCoord vc;
  vc = vcs.ictovc (imgcoordsys->lltoic (parent->pos));

  draw_cursor (&gd, (int) (vc.x + .5), (int) (vc.y + .5), 0.0);
  draw_scale (&gd, psd, (int) ps, (int) w, (int) h);


  save_pixbuf_to_stream(pb, iw, ih, f);
  gdk_pixbuf_finalize (pb);
#endif
}


void View::repaint (void)
{
  ViewCoordSystem vcs = vcs_from_view_and_canvas (this, canvas);
  int iw = (int) xv_get (canvas, WIN_WIDTH);
  int ih = (int) xv_get (canvas, WIN_HEIGHT);
  double w = (double) iw;
  ;
  double h = (double) ih;

  vcs.cursor.pos = parent->pos;
  vcs.cursor.heading = 0;


  GisDrawableX gdx (dpy, win, gc, iw, ih);

  int i, m, n = 0;

  for (i = 0; i < parent->nmaps; ++i) {
    if (usemaps[i] && parent->maps[i]->tiled) {
      n++;
      m = i;
    }
  }

  if (!n) {
    XFillRectangle (dpy, win, gc, 0, 0, iw, ih);
    return ;
  }

  n--;


  GdkPixbuf *b1 = 0, *b2 = 0;
  if (n) {
    b1 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, TILESIZE, TILESIZE);
    b2 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, False, 8, TILESIZE, TILESIZE);
  } else {
    b1 = 0;
    b2 = 0;
  }
  GisDrawableGdkPB gdg1 (b1);
  GisDrawableGdkPB gdg2 (b2);


  ImageCoord ic;
  for (double y = 0; y < h; y += TILESIZE) {
    for (double x = 0; x < w; x += TILESIZE) {
      ViewRect vr (x, y, TILESIZE, TILESIZE);

      if (n) {
        m = 0;
        for (i = 0; i < parent->nmaps; ++i) {
          if (usemaps[i] && parent->maps[i]->tiled) {
            if (!m) {
              gdg1.ox = (int) x;
              gdg1.oy = (int) y;
              parent->maps[i]->render (&gdg1, &vcs, &vr, RENDER_LAZY);
              m = 1;
            } else {
              gdg2.ox = (int) x;
              gdg2.oy = (int) y;
              parent->maps[i]->render (&gdg2, &vcs, &vr, RENDER_LAZY);

              gdk_pixbuf_composite (b2, b1, 0, 0, TILESIZE, TILESIZE, 0, 0, 1.0, 1.0, GDK_INTERP_NEAREST, 128);
            }
          }
        }

        gdx.render (b1, 0, 0, (int) (x + .5), (int) (y + .5), (int) TILESIZE, (int) TILESIZE, (int) x, (int) y);

      } else {
        parent->maps[m]->render (&gdx, &vcs, &vr, RENDER_LAZY);
      }


    }
  }

  if (n) {
    gdk_pixbuf_finalize (b2);
    gdk_pixbuf_finalize (b1);
  }

  ViewRect vr (0, 0, iw, ih);

  for (i = 0; i < parent->nmaps; ++i) {
    if (usemaps[i] && (!parent->maps[i]->tiled)) {
      parent->maps[i]->render (&gdx, &vcs, &vr, RENDER_LAZY);
    }
  }


}


static void
view_change_scale (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  double *sc = (double *) xv_get (pi, XV_KEY_DATA, SCALE_KEY);

  v->scale = *sc;

  printf ("New scale %.2f m/pixel\n", v->scale);

  canvas_repaint (v->canvas);
}

static void
view_change_scale_up (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  v->scale *= 1.2;
  printf ("New scale %.2f m/pixel\n", v->scale);
  canvas_repaint (v->canvas);
}

static void
view_change_scale_down (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  v->scale *= 0.8;
  printf ("New scale %.2f m/pixel\n", v->scale);
  canvas_repaint (v->canvas);
}

static void
maps_changed (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  int i;

  for (i = 0; i < v->parent->nmaps; ++i) {
    v->usemaps[i] = xv_get (v->mapmaps[i], PANEL_VALUE);
  }

  //xv_set(v->mapframe,XV_SHOW,FALSE,0);


  canvas_repaint (v->canvas);
}

static void
canvas_resize (Canvas c)
{
  View *v = (View *) xv_get (c, XV_KEY_DATA, VIEW_KEY);
  int w = xv_get (c, XV_WIDTH);
  int h = xv_get (c, XV_HEIGHT);

  if ((v->cw != w) || (v->ch != h)) {
    v->cw = w;
    v->ch = h;
    canvas_repaint (c);
  }
}


static Notify_value
canvas_event_proc (Canvas c, Event * e)
{
  View *v = (View *) xv_get (c, XV_KEY_DATA, VIEW_KEY);
  Rect *r;

  canvas_resize (c);

  if (!e)
    return NOTIFY_DONE;

  switch (event_action (e)) {
    case ACTION_SELECT:
      if (event_is_down (e)) {
        ViewCoordSystem vcs = vcs_from_view_and_canvas (v, c);
        ViewCoord vc ((double) event_x (e), (double) event_y (e));
        v->parent->pos = v->imgcoordsys->ictoll (vcs.vctoic (vc));
        v->parent->newpos ();
      }
      break;
    case LOC_MOVE:
      //XXX
      //Show where we are...

      break;

  }
  return NOTIFY_DONE;
}

static Notify_value view_destroy (Notify_client client, Destroy_status status)
{
  Frame frame = (Frame) client;
  View *v = (View *) xv_get (frame, XV_KEY_DATA, VIEW_KEY);


  switch (status) {
    case DESTROY_CLEANUP:
      if (v) {
        xv_set (v->frame, XV_KEY_DATA, VIEW_KEY, NULL, NULL);
        printf ("Destroying view at %08x\n", v);
        delete v;
      }
      return notify_next_destroy_func (client, status);
      break;
  }

  return NOTIFY_DONE;
}

View::~View ()
{
  View *v = (View *) xv_get (frame, XV_KEY_DATA, VIEW_KEY);
  if (v) {
    xv_set (v->frame, XV_KEY_DATA, VIEW_KEY, NULL, NULL);
    xv_destroy (frame);
  }
  XFreeGC (dpy, gc);
  delete[] usemaps;
  delete[] mapmaps;

  if (next) {
    next->prev = prev;
  } else {
    parent->tail = prev;
  }

  if (prev) {
    prev->next = next;
  } else {
    parent->head = next;
  }


}

static void print_print (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  char *where = (char *) panel_get_value(v->printwhere);
  int w = atoi((char *) panel_get_value(v->printxpx));
  int h = atoi((char *) panel_get_value(v->printypx));
  double r = atof((char *) panel_get_value(v->printres));
  double xl = atof((char *) panel_get_value(v->printxlen));
  double yl = atof((char *) panel_get_value(v->printylen));
  FILE *f;
  char buf[1024];



  if (!w)
    return ;
  if (!h)
    return ;

  if (!where)
    return ;

  sprintf(buf, where, 0.25*(300.0 / r), xl, yl);

  printf("Printing to %s\n", buf);

  if (buf[0] == '|') {
    f = popen(buf + 1, "w");
  } else {
    f = fopen(buf, "w");
  }

  if (!f)
    return ;

  v->print(f, w, h);

  xv_set(v->printstatus, PANEL_VALUE, "Waiting for command to terminate", 0);

  if (where[0] == '|') {
    pclose(f);
  } else {
    fclose(f);
  }

  xv_set(v->printstatus, PANEL_VALUE, "Done", 0);
  notify_dispatch();

}

static void print_recalculate (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  int change = (int) panel_get_value(v->printchange);
  char buf[80];
  double r = atof((char *) panel_get_value(v->printres));
  double xp = atof((char *) panel_get_value(v->printxpx));
  double yp = atof((char *) panel_get_value(v->printypx));
  double xl = atof((char *) panel_get_value(v->printxlen));
  double yl = atof((char *) panel_get_value(v->printylen));
  double xr = xp / xl;
  double yr = yp / yl;


  switch (change) {
    case 0:       //Resolution

      r = (xr > yr) ? xr : yr;

      sprintf(buf, "%.3f", r);

      xv_set(v->printres, PANEL_VALUE, buf, 0);

      //Fall through
    case 1:       //Pixels
      xp = xl * r;
      yp = yl * r;

      sprintf(buf, "%d", (int) xp);
      xv_set(v->printxpx, PANEL_VALUE, buf, 0);

      sprintf(buf, "%d", (int) yp);
      xv_set(v->printypx, PANEL_VALUE, buf, 0);


      break;
    case 2:       //Length
      if (!r)
        break;
      xl = xp / r;
      yl = yp / r;

      sprintf(buf, "%.2f", (int) xl);
      xv_set(v->printxlen, PANEL_VALUE, buf, 0);

      sprintf(buf, "%.2f", (int) yl);
      xv_set(v->printylen, PANEL_VALUE, buf, 0);

      break;
  }

}

static void print_setlengths(Panel_item pi, char *x, char *y)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  xv_set(v->printchange, PANEL_VALUE, (int) 1, 0);
  xv_set(v->printxlen, PANEL_VALUE, x, 0);
  xv_set(v->printylen, PANEL_VALUE, y, 0);
  print_recalculate(pi);
}

static void print_a4l(Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  print_setlengths(pi, "11.2", "7.8");
  xv_set(v->printwhere, PANEL_VALUE, "|pnmtops -turn -height 11.69 -width 8.27 -scale %.3f | lpr", 0);
}
static void print_a4p(Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  print_setlengths(pi, "7.7", "11.2");
  xv_set(v->printwhere, PANEL_VALUE, "|pnmtops -noturn -width 8.27 -height 11.69 -scale %.3f | lpr", 0);
}
static void print_done (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  xv_set (v->printframe, XV_SHOW, FALSE, 0);
}

static void maps_done (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  xv_set (v->mapframe, XV_SHOW, FALSE, 0);
}
static void print_proc(Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  xv_set (v->printframe, XV_SHOW, TRUE, 0);
}

static void what_proc (Panel_item pi)
{
  View *v = (View *) xv_get (pi, XV_KEY_DATA, VIEW_KEY);
  xv_set (v->mapframe, XV_SHOW, TRUE, 0);
}

View::View (Views * ipt, ImageCoordSystem * ics, double iscale, int *iusemaps)
{
  int i;
  parent = ipt;
  imgcoordsys = ics;
  scale = iscale;
  usemaps = iusemaps;

  prev = next = 0;


  frame = xv_create (0, FRAME, FRAME_LABEL, rcsid, XV_KEY_DATA, VIEW_KEY, this, WIN_SHOW, FALSE, WIN_WIDTH, VIEW_WIDTH + 10, WIN_HEIGHT, VIEW_HEIGHT + VIEW_PANEL_HEIGHT + 10, 0);

  dpy = (Display *) xv_get (frame, XV_DISPLAY);
  depth = xv_get (frame, WIN_DEPTH);



  panel = xv_create (frame, PANEL, XV_HEIGHT, VIEW_PANEL_HEIGHT, XV_WIDTH, VIEW_WIDTH, XV_KEY_DATA, VIEW_KEY, this, 0);

  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "What", PANEL_NOTIFY_PROC, what_proc, XV_KEY_DATA, VIEW_KEY, this, 0);


  char buf[100];
  static double scales[] = { 0.1, 0.25, 0.5, 1.0, 2.0, 2.5, 5.0, 10.0, 20.0, 50.0, 100.0, 200.0, 500.0, 1000.0 };
  int n = sizeof (scales) / sizeof (double);


  for (i = 0; i < n; ++i) {
    sprintf (buf, "%.2f", scales[i]);
    xv_create (panel, PANEL_BUTTON,
               PANEL_LABEL_STRING, buf,
               XV_KEY_DATA, VIEW_KEY, this,
               XV_KEY_DATA, SCALE_KEY, &scales[i],
               PANEL_NOTIFY_PROC, view_change_scale, 0);
  }

  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "Print", XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, print_proc, 0);


  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "+", XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, view_change_scale_up, 0);
  xv_create (panel, PANEL_BUTTON, PANEL_LABEL_STRING, "-", XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, view_change_scale_down, 0);


  mapframe = xv_create (frame, FRAME, XV_KEY_DATA, VIEW_KEY, this, 0);

  mappanel = xv_create (mapframe, PANEL, 0);

  string s = imgcoordsys->name ();

  xv_create (mappanel, PANEL_MESSAGE, PANEL_LABEL_STRING, s.c_str (), 0);


  mapmaps = new Panel_item[parent->nmaps];

  for (i = 0; i < parent->nmaps; ++i) {
    s = parent->maps[i]->name ();
    mapmaps[i] = xv_create (mappanel, PANEL_TOGGLE,
                            PANEL_CHOICE_STRINGS, s.c_str (), 0,
                            PANEL_VALUE, usemaps[i],
                            XV_KEY_DATA, VIEW_KEY, this,
                            PANEL_NOTIFY_PROC, maps_changed, 0);
  }

  xv_create (mappanel, PANEL_BUTTON, PANEL_LABEL_STRING, "Done",
             XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, maps_done, 0);
  window_fit (mappanel);
  window_fit (mapframe);


  printframe = xv_create (frame, FRAME, XV_KEY_DATA, VIEW_KEY, this, 0);

  printpanel = xv_create (printframe, PANEL, 0);

  printwhere = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                          "File/Command:", XV_KEY_DATA, VIEW_KEY, this,
                          PANEL_NOTIFY_PROC, print_print,
                          PANEL_VALUE, "map.ppm", 0);
  printchange = xv_create (printpanel, PANEL_CHOICE, PANEL_LABEL_STRING, "Auto adjust:",
                           PANEL_CHOICE_STRINGS, "Resolution", "Pixels", "Length", 0,
                           PANEL_VALUE, 1, 0);

  printres = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                        "Resolution:", XV_KEY_DATA, VIEW_KEY, this,
                        PANEL_VALUE, "200", PANEL_VALUE_DISPLAY_LENGTH, 12,
                        PANEL_NOTIFY_PROC, print_recalculate, 0);

  printxpx = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                        "Pixels X:", XV_KEY_DATA, VIEW_KEY, this,
                        PANEL_VALUE, "1024", PANEL_VALUE_DISPLAY_LENGTH, 12,
                        PANEL_NOTIFY_PROC, print_recalculate, 0);
  printypx = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                        "Y:", XV_KEY_DATA, VIEW_KEY, this,
                        PANEL_VALUE, "1024", PANEL_VALUE_DISPLAY_LENGTH, 12,
                        PANEL_NOTIFY_PROC, print_recalculate, 0);

  printxlen = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                         "Length X:", XV_KEY_DATA, VIEW_KEY, this,
                         PANEL_VALUE, "5.12", PANEL_VALUE_DISPLAY_LENGTH, 12,
                         PANEL_NOTIFY_PROC, print_recalculate, 0);
  printylen = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                         "Y:", XV_KEY_DATA, VIEW_KEY, this,
                         PANEL_VALUE, "5.12", PANEL_VALUE_DISPLAY_LENGTH, 12,
                         PANEL_NOTIFY_PROC, print_recalculate, 0);

  xv_create (printpanel, PANEL_BUTTON, PANEL_LABEL_STRING, "A4 Land",
             XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, print_a4l, 0);
  xv_create (printpanel, PANEL_BUTTON, PANEL_LABEL_STRING, "A4 Port",
             XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, print_a4p, 0);
  xv_create (printpanel, PANEL_BUTTON, PANEL_LABEL_STRING, "Print",
             XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, print_print, 0);
  xv_create (printpanel, PANEL_BUTTON, PANEL_LABEL_STRING, "Done",
             XV_KEY_DATA, VIEW_KEY, this, PANEL_NOTIFY_PROC, print_done, 0);

  printstatus = xv_create (printpanel, PANEL_TEXT, PANEL_LABEL_STRING,
                           "Status:", XV_KEY_DATA, VIEW_KEY, this,
                           PANEL_VALUE, "", PANEL_VALUE_DISPLAY_LENGTH, 60,
                           PANEL_READ_ONLY, TRUE, 0);

  window_fit (printpanel);
  window_fit (printframe);


  canvas = xv_create (frame, CANVAS,
                      XV_X, 0,
                      WIN_BELOW, panel,
                      WIN_WIDTH, VIEW_WIDTH, WIN_HEIGHT, VIEW_HEIGHT,
                      CANVAS_REPAINT_PROC, canvas_repaint,
                      WIN_EVENT_PROC, canvas_event_proc,
                      WIN_CONSUME_EVENTS, LOC_MOVE, LOC_DRAG,
                      WIN_MOUSE_BUTTONS, 0,
                      XV_KEY_DATA, VIEW_KEY, this,
                      CANVAS_AUTO_SHRINK, TRUE,
                      CANVAS_AUTO_EXPAND, TRUE, CANVAS_X_PAINT_WINDOW, TRUE,
                      //                     CANVAS_RESIZE_PROC, canvas_resize,
                      0);

  cw = xv_get (canvas, XV_WIDTH);
  ch = xv_get (canvas, XV_HEIGHT);

  xv_set (canvas_paint_window (canvas),
          WIN_EVENT_PROC, canvas_event_proc,
          WIN_CONSUME_EVENTS, LOC_MOVE, LOC_DRAG, WIN_MOUSE_BUTTONS, 0,
          XV_KEY_DATA, VIEW_KEY, this, 0);


  win = (XID) xv_get (canvas_paint_window (canvas), XV_XID);

  XGCValues gc_val;
  gc_val.line_width = 3;
  gc = XCreateGC (dpy, win, GCLineWidth, &gc_val);

  window_fit (frame);

  notify_interpose_destroy_func (frame, (Notify_func) view_destroy);

  xv_set (frame, XV_SHOW, TRUE, 0);
  printf ("Created new view\n");
}
