/*
* views.cc:
*
* Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
* All rights reserved.
*
*/

static char rcsid[] =
  "$Id: views.cc,v 1.12 2002/08/17 15:27:53 root Exp $";

/*
 * $Log: views.cc,v $
 * Revision 1.12  2002/08/17 15:27:53  root
 * #
 *
 * Revision 1.11  2002/08/15 15:30:59  root
 * #
 *
 * Revision 1.10  2002/08/15 11:41:06  root
 * #
 *
 * Revision 1.9  2002/06/21 17:33:06  root
 * #
 *
 * Revision 1.8  2001/12/16 17:56:08  root
 * #
 *
 * Revision 1.7  2001/12/16 17:44:02  root
 * #
 *
 * Revision 1.6  2001/12/16 17:42:43  root
 * #
 *
 * Revision 1.5  2001/12/16 17:33:58  root
 * #
 *
 * Revision 1.4  2001/12/16 17:32:38  root
 * #
 *
 * Revision 1.3  2001/12/06 21:04:21  root
 * #
 *
 * Revision 1.2  2001/12/06 17:47:19  root
 * #
 *
 * Revision 1.1  2001/12/06 17:47:08  root
 * #
 *
 */

#include "project.h"

static void new_view_notify (Panel_item pi)
{
  Views *vs = (Views *) xv_get (pi, XV_KEY_DATA, VIEWS_KEY);
  int m = (int) xv_get (pi, XV_KEY_DATA, MAP_KEY);
  vs->new_view (m);
}

static Notify_value views_destroy (Notify_client client, Destroy_status status)
{
  Frame frame = (Frame) client;
  Views *vs = (Views *) xv_get (frame, XV_KEY_DATA, VIEWS_KEY);


  switch (status) {
    case DESTROY_CLEANUP:
      if (vs) {
        xv_set (vs->frame, XV_KEY_DATA, VIEWS_KEY, NULL, NULL);
        delete vs;
      }
      return notify_next_destroy_func (client, status);
      break;
  }

  return NOTIFY_DONE;
}

static void place_notify (Panel_item pi)
{
  Views *vs = (Views *) xv_get (pi, XV_KEY_DATA, VIEWS_KEY);
  Place *p = (Place *) xv_get (pi, XV_KEY_DATA, PLACE_KEY);

  vs->pos.lat = p->lat;
  vs->pos.lon = p->lon;

  vs->newpos ();
}


static void cs_notify (Panel_item pi)
{
  Views *vs = (Views *) xv_get (pi, XV_KEY_DATA, VIEWS_KEY);
  CoordSystem *cs = (CoordSystem *) xv_get (pi, XV_KEY_DATA, CS_KEY);
  char *c = (char *) xv_get (pi, PANEL_VALUE);
  string s = "";

  while (*c) {
    if (isalnum (*c))
      s += *c;
    else if (*c == ',')
      s += *c;
    else if (*c == '.')
      s += *c;
    else if (*c == '-')
      s += *c;
    else if (*c == '\'')
      s += *c;
    else if (*c == '"')
      s += *c;
    c++;
  }

  vs->pos = cs->cstoll (s);

  vs->newpos ();

}

Views::Views (LatLon ipos, Map ** imaps, int inmaps, CoordSystem ** ics, int incs, Place ** ps, int nps)
{
  int i;
  string s;

  pos = ipos;
  maps = imaps;
  nmaps = inmaps;
  cs = ics;
  ncs = incs;
  head = tail = 0;

  frame = xv_create (0, FRAME, FRAME_LABEL, rcsid,
                     XV_KEY_DATA, VIEWS_KEY, this, 0);
  panel = xv_create (frame, PANEL, 0);

  csi = new Panel_item[ncs];
  //csi=(PanelItem *) malloc(sizeof(PanelItem)*ncs);


  for (i = 0; i < ncs; ++i) {
    s = cs[i]->name ();
    csi[i] = xv_create (panel, PANEL_TEXT,
                        PANEL_LABEL_STRING, s.c_str (),
                        XV_KEY_DATA, VIEWS_KEY, this,
                        XV_KEY_DATA, CS_KEY, cs[i],
                        PANEL_NOTIFY_PROC, cs_notify,
                        PANEL_VALUE_DISPLAY_LENGTH, 22, 0);
  }

  for (i = 0; i < nmaps; ++i) {
    if (maps[i]->preferedics) {
      s = maps[i]->name ();
      xv_create (panel, PANEL_BUTTON,
                 PANEL_LABEL_STRING, s.c_str (),
                 XV_KEY_DATA, VIEWS_KEY, this,
                 XV_KEY_DATA, MAP_KEY, i,
                 PANEL_NOTIFY_PROC, new_view_notify, 0);
    }
  }
  for (i = 0; i < nps; ++i) {
    xv_create (panel, PANEL_BUTTON,
               PANEL_LABEL_STRING, ps[i]->name.c_str (),
               XV_KEY_DATA, VIEWS_KEY, this,
               XV_KEY_DATA, PLACE_KEY, ps[i],
               PANEL_NOTIFY_PROC, place_notify, 0);
  }


  window_fit (panel);
  window_fit (frame);
  notify_interpose_destroy_func (frame, (Notify_func) views_destroy);
  xv_set (frame, XV_SHOW, TRUE, 0);
  newpos ();
}


void
Views::new_view (int mapnum)
{
  int *usemaps = new int[nmaps];

  for (int i = 0; i < nmaps; ++i)
    usemaps[i] = 0;

  usemaps[0] = 1;
  usemaps[1] = 1;
  usemaps[2] = 1;

  usemaps[mapnum] = 1;


  View *v = new View (this, maps[mapnum]->preferedics, maps[mapnum]->preferedscale, usemaps);

  if (head) {
    v->prev = 0;
    v->next = head;
    v->next->prev = v;
    head = v;
  } else {
    v->prev = 0;
    v->next = 0;
    head = v;
    tail = v;
  }


}

Views::~Views ()
{
  Views *vs = (Views *) xv_get (frame, XV_KEY_DATA, VIEWS_KEY);
  if (vs) {
    xv_set (vs->frame, XV_KEY_DATA, VIEWS_KEY, NULL, NULL);
    xv_destroy (frame);
  }
  while (head)
    delete head;
  delete[]csi;
  //free(csi);
}

void
Views::repaint_all (void)
{
  View *v = head;

  while (v) {
    v->repaint ();
    v = v->next;
  }

}



void
Views::newpos (void)
{
  int i;
  string s;

  for (i = 0; i < ncs; ++i) {
    s = cs[i]->lltocs (pos);
    xv_set (csi[i], PANEL_VALUE, s.c_str (), 0);
  }

  repaint_all ();
}
